from .read_multi_hist import read_multi_hist 
from ..line_size_over_time import line_size_over_time
from .multi_hist_complex_count import multi_hist_complex_count
from .multi_stack_hist_complex_count import multi_stack_hist_complex_count
from .multi_heatmap_complex_dist import multi_heatmap_complex_dist
from .multi_hist_3D_complex_dist import multi_hist_3D_complex_dist
from ..hist_to_csv import hist_to_csv
from ..hist_to_df import hist_to_df
from .frac_of_assemble import frac_of_assemble


class MultiHistogram ():
    """Multi Histogram object that holds all data from a mutli species histogram.dat 
    to be interpreting in many different ways
    """

    def __init__(self,FileName: str, FileNum: int, InitialTime: float, FinalTime: float, SpeciesList: list):
        """Will initilize the object by reading through the inputted file(s)

        Args:
            FileName (str): file location (relative) histogram.dat that will be read
            FileNum (int): Number of the total input files (file names should be [fileName]_1,[fileName]_2,...)
            InitialTime (float): The starting time. Must not be smaller / larger then times in file.
            FinalTime (float): The ending time. Must not be smaller / larger then times in file.
            SpeciesName (list): The names of the species that are in the multi-histogram file
                Ex: ['a','b']     
        """

        #Initilize variables
        self.full_hist = []
        self.FileName = FileName
        self.FileNum = FileNum
        self.InitialTime = InitialTime
        self.FinalTime = FinalTime
        self.SpeciesList = SpeciesList
        
        #setup file naming
        file_name_head = FileName.split('.')[0]
        file_name_tail = FileName.split('.')[1]

        for histogram_file_number in range(1, FileNum+1):
            
            #determining file name (if there are multiple or none)
            if FileNum == 1:
                temp_file_name = FileName
            else:
                temp_file_name = file_name_head + '_' + str(histogram_file_number) + '.' + file_name_tail
            
            #load in the file
            temp_hist = read_multi_hist(temp_file_name,SpeciesList)
            self.full_hist.append(temp_hist)


    ##Number of complexes over time (2d)
    def line_mean_complex_size(self, SpeciesName: str = "tot", ExcludeSize: int = 0, ShowFig: bool = True, SaveFig: bool = False, SaveVars: bool = False):
        """Creates graph of the mean number of species in a single complex molecule over a time period.

        Args:
            SpeciesName (str): The name of the species you want to examine. Should be in the .dat file.
            ExcludeSize (int, optional): Monomers in the complex that are smaller or equal to this number will not be included. 
            ShowFig (bool, optional): If the plot is shown. Defaults to True.
            SaveFig (bool, optional): If the plot is saved. Defaults to False.
            SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.

        Returns:
            graph. X-axis = time. Y
            -axis = mean number of species in a single complex molecule.
        """

        return line_size_over_time(Data = 1, full_hist = self.full_hist, FileNum = self.FileNum, InitialTime = self.InitialTime, FinalTime = self.FinalTime,
                SpeciesName = SpeciesName, ExcludeSize = ExcludeSize, SpeciesList = self.SpeciesList, ShowFig = ShowFig, SaveFig = SaveFig, SaveVars = SaveVars)


    def line_max_complex_size(self, SpeciesName: str = "tot", ExcludeSize: int = 0, ShowFig: bool = True, SaveFig: bool = False, SaveVars: bool = False):
            """Creates graph of the mean number of species in a single complex molecule over a time period.

            Args:
                SpeciesName (str): The name of the species you want to examine. Should be in the .dat file.
                ExcludeSize (int, optional): Monomers in the complex that are smaller or equal to this number will not be included. 
                ShowFig (bool, optional): If the plot is shown. Defaults to True.
                SaveFig (bool, optional): If the plot is saved. Defaults to False.
                SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.

            Returns:
                graph. X-axis = time. Y
                -axis = mean number of species in a single complex molecule.
            """

            return line_size_over_time(Data = 2, full_hist = self.full_hist, FileNum = self.FileNum, InitialTime = self.InitialTime, FinalTime = self.FinalTime,
                    SpeciesName = SpeciesName, ExcludeSize = ExcludeSize, SpeciesList = self.SpeciesList, ShowFig = ShowFig, SaveFig = SaveFig, SaveVars = SaveVars)


    ##Frequency of each complex size
    def hist_complex_count(self, BinNums: int = 10, ExcludeSize: int = 0, ShowFig: bool = True, SaveFig: bool = False, SaveVars: bool = False):
        """ Creates a 3D heatmap from a histogram.dat (multi-species) that shows distrubution of sizes of selected species.

        Args:
            BinNums (int, optional): The number of bins in the histogram. Default is 10.
            ExcludeSize (int, optional): The minimum value required to include a data point in the histogram. Default is 0.
            ShowFig (bool, optional): Whether to display the generated figures. Default is True.
            SaveFig (bool, optional): Whether to save the generated figures. Default is False.
            SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.

        Returns:
            3D Histogram. X-axis / Y-axis: the distribution of sizes of each specified species. Color: relative occurance of each complex.
        """

        return multi_hist_complex_count(self.FileName, self.FileNum, self.InitialTime, self.FinalTime, SpeciesList=self.SpeciesList,
                          BinNums=BinNums, ExcludeSize=ExcludeSize, ShowFig=ShowFig, SaveFig=SaveFig, SaveVars=SaveVars)



    ##Frequency of each kind of complex (stacked histogram)
    def stack_hist_complex_count(self,xAxis: str, DivideSpecies: str, DivideSize: int,
                        BarSize: int = 1, ExcludeSize: int = 0, ShowFig: bool = True, SaveFig: bool = False, SaveVars: bool = False):
        """Creates a stacked histogram from histogram.dat (multi-species) that shows the average number of each type of 
        complex species (based on protein composition) over the whole sim. 

        Args:
            xAxis (str): Species shown on X-axis.
            DivideSpecies (str): The name of the species that will be seperated by size.
            DivideSize (int): The value that separates the size of dissociate complexes. (only changes color of graph)
            BarSize (int, optional): The size of each data bar in the X-dimension. Defaults to 1.
            ExcludeSize (int, optional): Monomers in the complex that are smaller or equal to this number will not be included. 
            ShowFig (bool, optional): If the plot is shown. Defaults to True.
            SaveFig (bool, optional): If the plot is saved. Defaults to False.
            SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.

        Returns:
            Histogram. X-axis = size of selected species, Y-axis = average number of each corresponds.
        """

        return multi_stack_hist_complex_count(FullHist=self.full_hist, FileNum=self.FileNum, InitialTime= self.InitialTime, FinalTime = self.FinalTime, SpeciesList=self.SpeciesList, 
                                  xAxis=xAxis, DivideSpecies=DivideSpecies, DivideSize=DivideSize, BarSize=BarSize,ExcludeSize=ExcludeSize, ShowFig=ShowFig, SaveFig=SaveFig, SaveVars=SaveVars)


    ##Average count of each complex composition over the entire simulation time
    def heatmap_complex_dist(self,xAxis: str, yAxis: str, SpeciesList: list = [], xBarSize: int = 1, yBarSize: int = 1,
                    ShowFig: bool = True, ShowMean: bool = False, ShowStd: bool = False, SaveFig: bool = False, SaveVars: bool = False):
        """ Creates a 3D heatmap from a histogram.dat (multi-species) that shows distrubution of sizes of selected species.

        Args:
            xAxis (str): Species shown on X-axis.
            yAxis (str): Species shown on Y-axis.
            xBarSize (int, optional): The size of each data bar in the X-dimension. Defaults to 1.
            yBarSize (int, optional): The size of each data bar in the Y-dimension. Defaults to 1.
            ShowMean (bool, optional): If means will be shown in each box. Defaults to False.
            ShowStd (bool, optional): If std values will be shown in each box. Defaults to False.
            ShowFig (bool, optional): If the plot is shown. Defaults to True.
            SaveFig (bool, optional): If the plot is saved. Defaults to False.
            SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.

        Returns:
            3D Histogram. X-axis / Y-axis: the distribution of sizes of each specified species. Color: relative occurance of each complex.
        """

        return multi_heatmap_complex_dist(FileName=self.FileName, FileNum=self.FileNum, InitialTime=self.InitialTime, FinalTime=self.FinalTime, xAxis=xAxis, yAxis=yAxis,
                             SpeciesList=SpeciesList, xBarSize=xBarSize, yBarSize=yBarSize, ShowFig=ShowFig, ShowMean=ShowMean,
                             ShowStd=ShowStd, SaveFig=SaveFig, SaveVars=SaveVars)
    

    def hist_3D_complex_dist(self,xAxis: str, yAxis: str, SpeciesList: list = [], xBarSize: int = 1, yBarSize: int = 1,
                    ShowFig: bool = True, SaveFig: bool = False, SaveVars: bool = False):
        """ Creates a 3D heatmap from a histogram.dat (multi-species) that shows distrubution of sizes of selected species.

        Args:
            xAxis (str): Species shown on X-axis.
            yAxis (str): Species shown on Y-axis.
            xBarSize (int, optional): The size of each data bar in the X-dimension. Defaults to 1.
            yBarSize (int, optional): The size of each data bar in the Y-dimension. Defaults to 1.
            ShowMean (bool, optional): If means will be shown in each box. Defaults to False.
            ShowStd (bool, optional): If std values will be shown in each box. Defaults to False.
            ShowFig (bool, optional): If the plot is shown. Defaults to True.
            SaveFig (bool, optional): If the plot is saved. Defaults to False.
            SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.

        Returns:
            3D Histogram. X-axis / Y-axis: the distribution of sizes of each specified species. Color: relative occurance of each complex.
        """

        return multi_hist_3D_complex_dist(FileName=self.FileName, FileNum=self.FileNum, InitialTime=self.InitialTime, FinalTime=self.FinalTime, xAxis=xAxis, yAxis=yAxis,
                             SpeciesList=SpeciesList, xBarSize=xBarSize, yBarSize=yBarSize, ShowFig=ShowFig, SaveFig=SaveFig, SaveVars=SaveVars)


    def hist_to_csv(self,OpName: int = "histogram"):
        """Creates a .csv (spreadsheet) file from a histogram.dat file (multi-species)

        Args:
            OpName (str, Optional = "histogram"): what the outputted .csv file will be named

        Returns:
            histogram.csv file: Each row is a different time stamp (all times listed in column A). Each column is a different size of complex molecule (all sizes listed in row 1). Each box 
            is the number of that complex molecule at that time stamp.
        """  
        return hist_to_csv(self.full_hist, self.SpeciesList, OpName)

    def hist_to_df(self, OpName: int = "histogram", SaveCsv: bool = True):
        """Creates a pandas dataframe from a histogram.dat (multi-species)

        Args:
            OpName (str, Optional = "histogram"): what the outputted .csv file will be named
            SaveCsv (bool, optional): If a .csv file is saved as well. Defaults to True.

        Returns:
            pandas.df: Each row is a different time stamp (all times listed in column A). Each column is a different size of complex molecule (all sizes listed in row 1). Each box 
                is the number of that complex molecule at that time stamp.
        """
        return hist_to_df(self.full_hist, self.SpeciesList, OpName, SaveCsv)
    
    ##The Mysterious Frac of Assemble....
    def frac_of_assemble(self, Mol: int = "", Threshold: int = 2, ShowFig: bool = True, SaveFig: bool = False, SaveVars: bool = False):
        """Generates time dependence of the fraction of asssembled molecules from the histogram_complexes_time.dat in the input file within the specified size threshold.

        Args:
            Mol (str): The name of the molecule. Default is empty string.
            Threshold (int, optional): The minimum size considered to be assembled. Default is 2.
            ShowFig (bool, optional): Whether to display the generated figures. Default is True.
            SaveFig (bool, optional): Whether to save the generated figures. Default is False.
            SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.

        Returns:
            Line graph. X-axis: time stamps. Y-axis: average frac of assembly.
        """

        return frac_of_assemble(FileName=self.FileName, FileNum=self.FileNum, Mol = Mol, Threshold = Threshold, ShowFig=ShowFig,SaveFig=SaveFig, SaveVars=SaveVars)
    



from ..gen_platonic.distance import distance


def octa_vert_leg_reduce(COM: float, leg: float, sigma: float):
    """Reduces the length of an octagonal vertex leg based on a given center of mass (COM), leg vector, and sigma.

    The reduction of the leg length is calculated using the formula: leg_red = (leg - COM) * ratio + COM, where ratio
    is calculated as 1 minus half of the sigma divided by the distance between COM and leg, as given by the `distance`
    function from the `gen_platonic` module.

    Args:
        COM (float): The center of mass (COM) of the octagonal vertex.
        leg (float): The leg vector of the octagonal vertex.
        sigma (float): The sigma value used for reducing the length of the leg.

    Returns:
        list: A list containing the reduced leg vector (leg_red) of the octagonal vertex, with rounded values
        to 8 decimal places.

    Example:
        COM = [0, 0, 0]
        leg = [1, 1, 1]
        sigma = 0.5
        result = octa_vert_leg_reduce(COM, leg, sigma)
        print(result)
    """
    red_len = sigma/2
    ratio = 1 - red_len/distance(COM, leg)
    leg_red = []
    for i in range(0, 3):
        leg_red.append(round((leg[i] - COM[i])*ratio + COM[i], 8))
    return leg_red





from .cube_vert_COM_leg_gen import cube_vert_COM_leg_gen
from .cube_vert_leg_reduce import cube_vert_leg_reduce

def cube_vert_leg_reduce_coor_gen(radius: float, sigma: float):
    """Generates a list of reduced center of mass (COM) and leg coordinates for a cube vertex based on radius and sigma.

    This function generates a list of reduced center of mass (COM) and leg coordinates for a cube vertex based on the
    given radius and sigma values, using the `cube_vert_COM_leg_gen` and `cube_vert_leg_reduce` functions from the
    respective modules. The reduction is applied to each leg coordinate by calling the `cube_vert_leg_reduce` function
    with the appropriate arguments.

    Args:
        radius (float): The radius of the cube vertex, represented as a float value.
        sigma (float): The sigma value for the cube vertex, used to calculate the reduction ratio, represented as a float value.

    Returns:
        list: A list of lists, where each inner list contains four elements: the reduced center of mass (COM) coordinate
        and the reduced leg coordinates (leg1, leg2, and leg3) of a cube vertex, after applying the reduction ratio
        based on the given radius and sigma values.


    Example:
        cube_vert_leg_reduce_coor_gen(1.0, 0.2)
        # Generates a list of reduced center of mass (COM) and leg coordinates for a cube vertex with a radius of 1.0 and
        # a sigma value of 0.2.
        # Returns a list of lists, where each inner list contains four elements: the reduced center of mass (COM) coordinate
        # and the reduced leg coordinates (leg1, leg2, and leg3) of a cube vertex, after applying the reduction ratio
        # based on the given radius and sigma values.
    """
    
    COM_leg_list = cube_vert_COM_leg_gen(radius)
    COM_leg_red_list = []
    for elements in COM_leg_list:
        temp_list = []
        temp_list.append(elements[0])
        i = 1
        while i <= 3:
            temp_list.append(cube_vert_leg_reduce(
                elements[0], elements[i], sigma))
            i += 1
        COM_leg_red_list.append(temp_list)
    return COM_leg_red_list





import os
import shutil
from pathlib import Path
from .merge_files import merge_files


def merge_simulation_results():
    """
    Merge the results of restart simulations.

    This function merges the content of PDB and RESTARTS folders from each numbered
    folder and their corresponding restart folders, and combines the .dat files.

    The base folder includes the folders named 1, 2, 3,… for each simulation run;
    each folder has the following structure, (d) means it is a directory:
    —1 (d)
        —PDB (d)
        —RESTARTS (d)
        —bound_pair_time.dat
        —copy_numbers_time.dat
        —event_counters_time.dat
        —histogram_complexes_time.dat
        —transition_matrix_time.dat
        —restart#### (d)
            —PDB (d)
            —RESTARTS (d)
            —bound_pair_time.dat
            —copy_numbers_time.dat
            —event_counters_time.dat
            —histogram_complexes_time.dat
            —transition_matrix_time.dat
        —restart#### (d)
            —PDB (d)
            —RESTARTS (d)
            —bound_pair_time.dat
            —copy_numbers_time.dat
            —event_counters_time.dat
            —histogram_complexes_time.dat
            —transition_matrix_time.dat

    output are under folders 1,2,3...
    """

    numbered_folders = [f for f in os.listdir('.') if f.isdigit()]

    for folder in numbered_folders:
        folder_path = os.path.join('.', folder)

        # Find all restart folders
        restart_folders = [d for d in os.listdir(folder_path) if d.startswith("restart") and os.path.isdir(os.path.join(folder_path, d))]


        # Sort restart folders by the timestamp
        sorted_restart_folders = sorted(restart_folders, key=lambda x: int(x[7:]))

        # Move PDB files from the restart to the root of each simulation folder
        for restart_folder in sorted_restart_folders:
            pdb_source_path = os.path.join(folder_path, restart_folder, "PDB")
            pdb_destination_path = os.path.join(folder_path, "PDB")
            for pdb_file in os.listdir(pdb_source_path):
                source_file = os.path.join(pdb_source_path, pdb_file)
                destination_file = os.path.join(pdb_destination_path, pdb_file)
                if not os.path.exists(destination_file):
                    shutil.move(source_file, pdb_destination_path)

        # Move restart files from the restart to the root of each simulation folder
            restarts_source_path = os.path.join(folder_path, restart_folder, "RESTARTS")
            restarts_destination_path = os.path.join(folder_path, "RESTARTS")
            for restart_file in os.listdir(restarts_source_path):
                source_file = os.path.join(restarts_source_path, restart_file)
                destination_file = os.path.join(restarts_destination_path, restart_file)
                if not os.path.exists(destination_file):
                    shutil.move(source_file, restarts_destination_path)

        # Copy the restart.dat file from the latest restart folder to the main folder
        if sorted_restart_folders:
            latest_restart_folder = sorted_restart_folders[-1]
            try:
                shutil.copy(os.path.join(folder_path, latest_restart_folder, "restart.dat"), folder_path)
            except:
                pass

        # Merge .dat files
        dat_files = ["bound_pair_time.dat", "copy_numbers_time.dat", "event_counters_time.dat",
                     "histogram_complexes_time.dat", "transition_matrix_time.dat"]

        for dat_file in dat_files:
            file_type = dat_file.split("_")[0]

            # Merge .dat files from the sorted restart folders
            for restart_folder in sorted_restart_folders:
                merge_files(os.path.join(folder_path, dat_file), os.path.join(folder_path, restart_folder, dat_file), file_type)

        # Remove all the restart folders
        for restart_folder in restart_folders:
            shutil.rmtree(os.path.join(folder_path, restart_folder))



def read_PDB(file_name, drop_COM):
    """Converts a PDB file to an array.

    Args:
        file_name (str): Name of the PDB file to be read.
        drop_COM (bool): Whether to drop lines with 'COM' as the 'Cite_Name' value.

    Returns:
        numpy array: array that stores all of the proteins info
            - [i] = each row, different protein site
            - [i][i] = each column (in a specific row) find the index of the correct colum with the dictionary
                - Ex: site_dict['Protein_Num'] >> 0
        dictionary: dictionary that stores the index of each column
        dictionary: stores the protein type for each protein number
        array: stores each line, and the important parts of it for use in writing    
    
    """

    site_array = []
    site_dict = {'Protein_Num':0,'Protein_Name':1,'Site_Name':2,'x_coord':3,'y_coord':4,'z_coord':5}
    num_name_dict = {}
    main_pdb_lst = []

    #opens the file, goes through each line, and stores it into the array. Also, creates list of each line for use later.
    with open(file_name, 'r') as file:
        
        for line in file.readlines():
            line_lst = line.split(' ')
            
            #if the line discribes a site, take info from it and put it into site_array
            if line_lst[0] == 'ATOM':
                info = []
                info = [element.strip('\n') for element in line_lst if element !=""] #removes white space from line
                    
                if (drop_COM and info[2] != 'COM') or not drop_COM: #if drop_COM is true, dont include COM. if it is false, include everything.

                    site_array.append([])
                    index = len(site_array) - 1

                    site_array[index].append(int(info[4]))
                    site_array[index].append(info[3])
                    site_array[index].append(info[2])
                    site_array[index].append(float(info[5]))
                    site_array[index].append(float(info[6]))
                    site_array[index].append(float(info[7]))
                    
                    num_name_dict[info[4]] = info[3]

            #Creates list of every line
                main_pdb_lst.append([int(info[4]),line])
            else:
                main_pdb_lst.append(['header',line])
                    

                        

    return site_array,site_dict,num_name_dict,main_pdb_lst





import os
import importlib

from .gen import *

# Get the directory of the current __init__.py file
current_directory = os.path.dirname(__file__)

# Iterate through all files in the current_directory
for filename in os.listdir(current_directory):
    # Check if the file is a Python file (ends with .py) and is not the current __init__.py
    if filename.endswith("nerdss.py") and not filename.startswith("__init__"):
        # Remove the .py extension from the filename to get the module name
        module_name = filename[:-3]

        # Import the module using importlib.import_module and add it to the globals dictionary
        module = importlib.import_module(f".{module_name}", package=__name__)
        globals().update(
            {n: getattr(module, n) for n in module.__all__}
            if hasattr(module, "__all__")
            else {k: v for k, v in module.__dict__.items() if not k.startswith("_")}
        )



from ..gen_platonic.distance import distance


def icos_vert_check_dis(cen: float, COM: float, lg1: float, lg2: float, lg3: float, lg4: float, lg5: float):
    """Check distances between a center point and other points.

    Args:
        cen (float): The center point.
        COM (float): The center of mass point.
        lg1 (float): Point 1.
        lg2 (float): Point 2.
        lg3 (float): Point 3.
        lg4 (float): Point 4.
        lg5 (float): Point 5.

    Returns:
        tuple: A tuple containing the distances between the center point and other points.
    """
    dis1 = round(distance(cen, lg1), 8)
    dis2 = round(distance(cen, lg2), 8)
    dis3 = round(distance(cen, lg3), 8)
    dis4 = round(distance(cen, lg4), 8)
    dis5 = round(distance(cen, lg5), 8)
    dis_ = round(distance(COM, cen), 8)
    if dis1 == dis2 and dis1 == dis3 and dis1 == dis4 and dis1 == dis5:
        return dis1, dis_
    else:
        return dis1, dis_





import os
import numpy as np

def save_multiple_arrays_to_file(arrays, file_name=None, file_path=None):
    """
    Save multiple NumPy arrays of numbers to a .dat file. Each array is stored on one column

    This function takes a list of arrays as input, saves it to a .dat file
    with each array on a column, and allows for customization of the file name and file
    path. If no file name is provided, the default name is "my_arrays". If no file path is
    provided, the default path is the current folder.

    Parameters
    ----------
    arrays : list of numpy.arrays
        The list of arrays to be saved.
    file_name : str, optional
        The name of the file, by default "my_arrays".
    file_path : str, optional
        The abs path of the file, by default the current folder.

    Returns
    -------
    None
    """
    if not isinstance(arrays, (list, np.ndarray)):
        raise ValueError("The input variable must be a list or a NumPy array of numbers.")
    
    if isinstance(arrays, np.ndarray):
        arrays = arrays.tolist()
    
    if not all(isinstance(item, np.ndarray) for item in arrays):
        raise ValueError("All elements in the list or array must be np.array.")
    
    if file_name is None:
        file_name = "my_arrays" + '.dat'
    else:
        file_name = file_name + '.dat'

    if file_path is None:
        file_path = os.getcwd()
    else:
        file_path = os.path.abspath(file_path)

    file_full_path = os.path.join(file_path, file_name)

    # combine the arrays in the list arrays vertically to create a 2D array
    data = np.vstack(arrays).T

    # save the data to a data file named file_full_path
    np.savetxt(file_full_path, data, delimiter=',')



import numpy as np
import os
import matplotlib.pyplot as plt
from ...file_managment.save_vars_to_file import save_vars_to_file


def frac_of_assemble(FileName: str = 'histogram_complexes_time.dat', FileNum: int = 1, 
               Mol: str = '', Threshold: int = 2, ShowFig: bool = True, SaveFig: bool = False, SaveVars: bool = False) -> tuple:
    """
    Generate time dependence of the fraction of asssembled molecules from the histogram_complexes_time.dat in the input file within the specified size threshold.

    Args:
        FileName (str): The name of the input file. Default is 'histogram_complexes_time.dat'
        FileNum (int): The number of files to read. Default is 1.
        Mol (str): The name of the molecule. Default is empty string.
        Threshold (int, optional): The minimum size considered to be assembled. Default is 2.
        ShowFig (bool, optional): Whether to display the generated figures. Default is True.
        SaveFig (bool, optional): Whether to save the generated figures. Default is False.
        SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.

    Returns:
        the np.arrays of time, means, and standard deviations for the fraction.
    """

    FileName_head, FileName_tail = os.path.splitext(FileName)

    time_arrays = []
    frac_arrays = []
     
    for i in range(1, FileNum+1):
        # check if there is a folder named '1' in the current folder, if there is, set the temp_FileName to ./i/FileName, then check if ./i/FileName exited. if not exited, throw out a file not exit error.
        # else if FileNum is 1, check if there is file named FileName in the current folder, if there is, set the temp_FileName to FileName; if FileNum > 1, check if there is afile name FileName_head + '_' + str(i) + '.' + FileName_tail, if it is, set as the temp_FileName
        # else throw a file not found error
        temp_FileName = ""
        if os.path.isdir(str(i)):
            temp_FileName = os.path.join(str(i), FileName)
            if not os.path.isfile(temp_FileName):
                raise FileNotFoundError(f"File {temp_FileName} does not exist.")
        elif FileNum == 1:
            if os.path.isfile(FileName):
                temp_FileName = FileName
            else:
                raise FileNotFoundError(f"File {FileName} does not exist.")
        else:
            temp_FileName = f"{FileName_head}_{i}.{FileName_tail}"
            if not os.path.isfile(temp_FileName):
                raise FileNotFoundError(f"File {temp_FileName} does not exist.")

        time_list = []
        frac_list = []
        assembled_num = 0
        non_assembled_num = 0
        # Open file and read line-by-line
        with open(temp_FileName, 'r') as file:
            found_startTime = False

            for line in file:
                # Check if line contains time information
                if line.startswith('Time (s):'):
                    if found_startTime:
                        # store the previous step result
                        frac = assembled_num / (assembled_num + non_assembled_num)
                        time_list.append(time)
                        frac_list.append(frac)
                        assembled_num = 0
                        non_assembled_num = 0

                    time = float(line.split()[-1])   # Extract time value from line
                    found_startTime = True

                else:   # Line contains complex count information
                    # Extract count and complex information
                    count, complexInfoStr = line.strip().split('\t')
                    count = int(count)
                    # Split complex information into individual items
                    complexItems = complexInfoStr.rstrip('.').split('. ')
                    size = 0
                    for item in complexItems:
                        # Extract species name and count from complex information
                        speciesName, numStr = item.split(': ')
                        num = int(numStr)
                        if speciesName == Mol and num >= Threshold:
                            assembled_num = assembled_num + num * count
                        elif speciesName == Mol and num < Threshold:
                            non_assembled_num = non_assembled_num + num * count
            frac = assembled_num / (assembled_num + non_assembled_num)
            time_list.append(time)
            frac_list.append(frac)
        time_array = np.array(time_list)
        frac_array = np.array(frac_list)
        time_arrays.append(time_array)
        frac_arrays.append(frac_array)

    # Find the length of the shortest time_array and frac_array
    min_length = min([len(time_array) for time_array in time_arrays])

    # Truncate all time_arrays and frac_arrays to the shortest length
    truncated_time_arrays = [time_array[:min_length] for time_array in time_arrays]
    truncated_frac_arrays = [frac_array[:min_length] for frac_array in frac_arrays]

    # Stack the truncated time_arrays and frac_arrays
    stacked_time_arrays = np.stack(truncated_time_arrays)
    stacked_frac_arrays = np.stack(truncated_frac_arrays)

    # Compute the average and standard deviation along the first axis
    average_time_array = np.mean(stacked_time_arrays, axis=0)
    average_frac_array = np.mean(stacked_frac_arrays, axis=0)
    std_frac_array = np.std(stacked_frac_arrays, axis=0)

    #output variables
    if SaveVars:
        save_vars_to_file({"time_stamp":average_time_array, "frac_assembled_monos":average_frac_array, "std":std_frac_array})

    # Plot the average_frac_array versus average_time_array with an error band
    plt.plot(average_time_array, average_frac_array)
    plt.fill_between(average_time_array, average_frac_array - std_frac_array, average_frac_array + std_frac_array, alpha=0.2)
    plt.xlabel('Time (s)')
    plt.ylabel('Average Frac of Assembly')
    if SaveFig:
        plt.SaveFig('FracofAssembly.png', dpi=300)
    if ShowFig:
        plt.show()
    
    return average_time_array, average_frac_array, std_frac_array


import numpy as np


def read_multi_hist(FileName: str, SpeciesList: list):
    """Read a multi-species histogram from a file and return a list of time steps
    and species counts.

    Args:
        FileName (str): The name of the input file to read.
        SpeciesList (list): A list of species names corresponding to the columns in the input file.

    Returns:
        hist_list (List): A list of time steps and species counts. Each element of the list is a
            NumPy array with shape (N, M+1), where N is the number of time steps and
            M is the number of species in SpeciesList. The first column contains the
            time step and the remaining columns contain the counts for each species.

    Raises:
        ValueError: If any of the species names in the input file are not found in SpeciesList.

    Examples:
        >>> SpeciesList = ['A', 'B', 'C', 'D']
        >>> read_multi_hist('histogram_complexes_time.dat', SpeciesList)

    Notes:
        The input file must have the following format:

        Time (s): <time>
        <complex_count><tab><species 1>: <num_of_species_1_in_complex> . <species 2>: <num_of_species_2_in_complex> . ...
        <complex_count><tab><species 1>: <num_of_species_1_in_complex> . <species 2>: <num_of_species_2_in_complex> . ...
        ...

        where <time> is the time step in seconds and <count> is the number of counts
        for each species at that time step. Each line after the first should contain
        one time step and its corresponding species counts, with species counts
        separated by periods.
    """
    speciesListArray = np.array(
        SpeciesList)  # Convert SpeciesList to a NumPy array
    histogramList = []  # Initialize empty list to store histograms
    # Initialize empty list to store complex counts for each time step
    timeStepComplexCounts = []
    previous_time = -1 # to ensure that multiple lines are not added for the same timestep

    # Create dictionary to store species indices
    speciesIndexDict = {speciesName: index for index, speciesName in enumerate(speciesListArray)}

    # Open file and read line-by-line
    with open(FileName, 'r') as file:
        for line in file:
            
            # Check if line contains time information
            if line.startswith('Time (s):'):
                time = float(line.split()[-1])   # Extract time value from line
                # Start new time step with current time value
                timeStepComplexCounts = [time]
            
            else:   # Line contains complex count information
                # Extract count and complex information
                count, complexInfoStr = line.strip().split('\t')
                # Split complex information into individual items
                complexItems = complexInfoStr.rstrip('.').split('. ')
                # Initialize array to store complex counts
                complexCounts = np.zeros(len(speciesListArray) + 1, dtype=int)
                for item in complexItems:
                    # Extract species name and count from complex information
                    speciesName, numStr = item.split(': ')
                    num = int(numStr)
                    # Get index of species from dictionary
                    if speciesName in speciesIndexDict:
                        index = speciesIndexDict[speciesName]
                    else:
                        raise Exception(f"{speciesName} was not included in SpeciesList")
                    # Increment count for species in complexCounts
                    complexCounts[index] += num
                # Add total count to last element of complexCounts
                complexCounts[-1] += int(count)
                # Add complexCounts to current time step
                timeStepComplexCounts.append(complexCounts)
            # Append final time step to histogram list
            if timeStepComplexCounts[0] != previous_time:
                histogramList.append(timeStepComplexCounts)
                previous_time = timeStepComplexCounts[0]
    

    return histogramList   # Return list of histograms for each time step





from .dode_face_dodecahedron_coord import dode_face_dodecahedron_coord
from .dode_face_COM_leg_coor import dode_face_COM_leg_coor


def dode_face_COM_leg_list_gen(radius: float):
    """Generate the Center of Mass (COM) and leg coordinates of 12 faces of a dodecahedron.

    Args:
        radius (float): The radius of the dodecahedron.

    Returns:
        list: A list containing the COM and leg coordinates of 12 faces as a large list.

    """
    
    # generate all COM and leg coords of 12 faces as a large list
    coord = dode_face_dodecahedron_coord(radius)
    COM_leg_list = []
    COM_leg_list.append(dode_face_COM_leg_coor(
        coord[6], coord[18], coord[2], coord[14], coord[4]))
    COM_leg_list.append(dode_face_COM_leg_coor(
        coord[6], coord[4], coord[12], coord[0], coord[16]))
    COM_leg_list.append(dode_face_COM_leg_coor(
        coord[4], coord[14], coord[9], coord[8], coord[12]))
    COM_leg_list.append(dode_face_COM_leg_coor(
        coord[6], coord[18], coord[11], coord[10], coord[16]))
    COM_leg_list.append(dode_face_COM_leg_coor(
        coord[14], coord[2], coord[3], coord[15], coord[9]))
    COM_leg_list.append(dode_face_COM_leg_coor(
        coord[18], coord[11], coord[19], coord[3], coord[2]))
    COM_leg_list.append(dode_face_COM_leg_coor(
        coord[16], coord[10], coord[17], coord[1], coord[0]))
    COM_leg_list.append(dode_face_COM_leg_coor(
        coord[12], coord[0], coord[1], coord[13], coord[8]))
    COM_leg_list.append(dode_face_COM_leg_coor(
        coord[7], coord[17], coord[10], coord[11], coord[19]))
    COM_leg_list.append(dode_face_COM_leg_coor(
        coord[5], coord[13], coord[8], coord[9], coord[15]))
    COM_leg_list.append(dode_face_COM_leg_coor(
        coord[3], coord[19], coord[7], coord[5], coord[15]))
    COM_leg_list.append(dode_face_COM_leg_coor(
        coord[1], coord[17], coord[7], coord[5], coord[13]))
    return COM_leg_list





from .read_restart import read_restart
from .write_pdb import write_pdb
import numpy as np

def single_locate_position_restart(FileNamePdb, ComplexSize, FileNameRestart='restart.dat'):
    """ Reads a restart.dat file and a PDB file, identifies protein complexes of a certain size, creates a new PDB file
    containing only the proteins corresponding to those complexes, and writes the new PDB file.

    Args:
        FileNamePdb (str): the name of the input PDB file
        ComplexSize (int): the size of protein complexes to be located
        FileNameRestart (str, optional): the name of the input restart.dat file (default is 'restart.dat')

    Returns:
        output_file.pdb: holds all of the proteins that were in a complex of a certain size

    """

    #read restart file
    print('Reading restart.dat...')
    complex_lst = read_restart(FileNameRestart)
    print('Reading files complete!')
    
    #find which complexes have the correct size
    protein_remain = []
    for complex in complex_lst:
        if len(complex) == ComplexSize:
            print(complex)
            print(len(complex))
            protein_remain.append(complex)
    
    #flatten list as it is a list of lists (list of complexes)
    protein_remain_flat = np.array(protein_remain).flatten()
    
    #write new PDB
    write_pdb(FileNamePdb, protein_remain_flat)
    print('PDB writing complete!(named as output_file.pdb)')
    return 0


from .icos_face_vert_coord import icos_face_vert_coord
from .icos_face_COM_coord import icos_face_COM_coord


def icos_face_COM_list_gen(radius: float):
    """Generates a list of coordinates representing the centers of mass (COM) of the faces of an icosahedron,
    given the radius of the icosahedron.

    Args:
        radius (float): The radius of the icosahedron.

    Returns:
        list: A list of 20 COM coordinates, each representing the center of mass of a face of the icosahedron.
    """
    coord = icos_face_vert_coord(radius)
    COM_list = []
    COM_list.append(icos_face_COM_coord(coord[0], coord[2], coord[8]))
    COM_list.append(icos_face_COM_coord(coord[0], coord[8], coord[4]))
    COM_list.append(icos_face_COM_coord(coord[0], coord[4], coord[6]))
    COM_list.append(icos_face_COM_coord(coord[0], coord[6], coord[10]))
    COM_list.append(icos_face_COM_coord(coord[0], coord[10], coord[2]))
    COM_list.append(icos_face_COM_coord(coord[3], coord[7], coord[5]))
    COM_list.append(icos_face_COM_coord(coord[3], coord[5], coord[9]))
    COM_list.append(icos_face_COM_coord(coord[3], coord[9], coord[1]))
    COM_list.append(icos_face_COM_coord(coord[3], coord[1], coord[11]))
    COM_list.append(icos_face_COM_coord(coord[3], coord[11], coord[7]))
    COM_list.append(icos_face_COM_coord(coord[7], coord[2], coord[5]))
    COM_list.append(icos_face_COM_coord(coord[2], coord[5], coord[8]))
    COM_list.append(icos_face_COM_coord(coord[5], coord[8], coord[9]))
    COM_list.append(icos_face_COM_coord(coord[8], coord[9], coord[4]))
    COM_list.append(icos_face_COM_coord(coord[9], coord[4], coord[1]))
    COM_list.append(icos_face_COM_coord(coord[4], coord[1], coord[6]))
    COM_list.append(icos_face_COM_coord(coord[1], coord[6], coord[11]))
    COM_list.append(icos_face_COM_coord(coord[6], coord[11], coord[10]))
    COM_list.append(icos_face_COM_coord(coord[11], coord[10], coord[7]))
    COM_list.append(icos_face_COM_coord(coord[10], coord[7], coord[2]))
    return COM_list





# take in the pdb file and output the pdb file with the corse-grained interfaces added

import numpy as np
import os


class Chain:
    def __init__(self, name, index):
        self.name = name
        self.index = index
        self.residues = []
        self.COM = None
        self.interfaces = []
        self.atoms = []


class Residue:
    def __init__(self, index):
        self.index = index
        self.atoms = []
        self.chain = -1
        self.CA = -1
        self.isDNA = False


class Coord:
    def __init__(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z

    def distanceSquare(self, other):
        dx = self.x - other.x
        dy = self.y - other.y
        dz = self.z - other.z
        dsquare = dx * dx + dy * dy + dz * dz
        return dsquare

    # define the - between Coord and list, return list
    def __sub__(self, other):
        return [self.x - other[0], self.y - other[1], self.z - other[2]]


class Atom:
    def __init__(
        self, index, atomType, residueType, chainName, chainIndex, residueIndex, coord
    ):
        self.index = index
        self.atomType = atomType
        self.residueType = residueType
        self.chainName = chainName
        self.chainIndex = chainIndex
        self.resideIndex = residueIndex
        self.coord = coord


class Interface:
    def __init__(self, index, partner):
        self.index = index
        self.partner = partner
        self.coord = None
        self.name = None
        self.chain = -1
        self.energy = 0.0
        self.atomsNum = 0
        self.residuesNum = 0
        self.size = 0


class Molecule:
    def __init__(self, name):
        self.name = name
        self.COM = None
        self.COMVisual = None
        self.interfaces = []
        self.radius = 0.0
        self.D = 0.0
        self.Dr = 0.0


class Intf:
    def __init__(self, name, positions):
        self.name = name
        self.positions = positions
        self.positionsVisual = None
        self.states = []


class State:
    def __init__(self, name):
        self.name = name


class Reaction:
    def __init__(self, name, reactants, products):
        self.name = name
        self.reactants = reactants
        self.products = products
        self.reverse = True
        self.onRate = 10.0
        self.offRate = 1.0
        self.angles = []
        self.norm1 = []
        self.norm2 = []
        self.sigma = 0.0
        self.energy = 0.0
        self.size = 0


def cg(filename, cutoff=3.5, thresholdResidue=3, stretchFactor=1.0):
    energyTable = {
        "CYS-CYS": 5.44,
        "CYS-MET": 4.99,
        "CYS-PHE": 5.80,
        "CYS-ILE": 5.50,
        "CYS-LEU": 5.83,
        "CYS-VAL": 4.96,
        "CYS-TRP": 4.95,
        "CYS-TYR": 4.16,
        "CYS-ALA": 3.57,
        "CYS-GLY": 3.16,
        "CYS-THR": 3.11,
        "CYS-SER": 2.86,
        "CYS-ASN": 2.59,
        "CYS-GLN": 2.85,
        "CYS-ASP": 2.41,
        "CYS-GLU": 2.27,
        "CYS-HIS": 3.60,
        "CYS-ARG": 2.57,
        "CYS-LYS": 1.95,
        "CYS-PRO": 3.07,
        "MET-MET": 5.46,
        "MET-PHE": 6.56,
        "MET-ILE": 6.02,
        "MET-LEU": 6.41,
        "MET-VAL": 5.32,
        "MET-TRP": 5.55,
        "MET-TYR": 4.91,
        "MET-ALA": 3.94,
        "MET-GLY": 3.39,
        "MET-THR": 3.51,
        "MET-SER": 3.03,
        "MET-ASN": 2.95,
        "MET-GLN": 3.30,
        "MET-ASP": 2.57,
        "MET-GLU": 2.89,
        "MET-HIS": 3.98,
        "MET-ARG": 3.12,
        "MET-LYS": 2.48,
        "MET-PRO": 3.56,
        "PHE-PHE": 7.26,
        "PHE-ILE": 6.84,
        "PHE-LEU": 7.28,
        "PHE-VAL": 6.29,
        "PHE-TRP": 6.16,
        "PHE-TYR": 5.66,
        "PHE-ALA": 4.81,
        "PHE-GLY": 4.13,
        "PHE-THR": 4.28,
        "PHE-SER": 4.02,
        "PHE-ASN": 3.75,
        "PHE-GLN": 4.10,
        "PHE-ASP": 3.48,
        "PHE-GLU": 3.56,
        "PHE-HIS": 4.77,
        "PHE-ARG": 3.98,
        "PHE-LYS": 3.36,
        "PHE-PRO": 4.25,
        "ILE-ILE": 6.54,
        "ILE-LEU": 7.04,
        "ILE-VAL": 6.05,
        "ILE-TRP": 5.78,
        "ILE-TYR": 5.25,
        "ILE-ALA": 4.58,
        "ILE-GLY": 3.78,
        "ILE-THR": 4.03,
        "ILE-SER": 3.52,
        "ILE-ASN": 3.24,
        "ILE-GLN": 3.67,
        "ILE-ASP": 3.17,
        "ILE-GLU": 3.27,
        "ILE-HIS": 4.14,
        "ILE-ARG": 3.63,
        "ILE-LYS": 3.01,
        "ILE-PRO": 3.76,
        "LEU-LEU": 7.37,
        "LEU-VAL": 6.48,
        "LEU-TRP": 6.14,
        "LEU-TYR": 5.67,
        "LEU-ALA": 4.91,
        "LEU-GLY": 4.16,
        "LEU-THR": 4.34,
        "LEU-SER": 3.92,
        "LEU-ASN": 3.74,
        "LEU-GLN": 4.04,
        "LEU-ASP": 3.40,
        "LEU-GLU": 3.59,
        "LEU-HIS": 4.54,
        "LEU-ARG": 4.03,
        "LEU-LYS": 3.37,
        "LEU-PRO": 4.20,
        "VAL-VAL": 5.52,
        "VAL-TRP": 5.18,
        "VAL-TYR": 4.62,
        "VAL-ALA": 4.04,
        "VAL-GLY": 3.38,
        "VAL-THR": 3.46,
        "VAL-SER": 3.05,
        "VAL-ASN": 2.83,
        "VAL-GLN": 3.07,
        "VAL-ASP": 2.48,
        "VAL-GLU": 2.67,
        "VAL-HIS": 3.58,
        "VAL-ARG": 3.07,
        "VAL-LYS": 2.49,
        "VAL-PRO": 3.32,
        "TRP-TRP": 5.06,
        "TRP-TYR": 4.66,
        "TRP-ALA": 3.82,
        "TRP-GLY": 3.42,
        "TRP-THR": 3.22,
        "TRP-SER": 2.99,
        "TRP-ASN": 3.07,
        "TRP-GLN": 3.11,
        "TRP-ASP": 2.84,
        "TRP-GLU": 2.99,
        "TRP-HIS": 3.98,
        "TRP-ARG": 3.41,
        "TRP-LYS": 2.69,
        "TRP-PRO": 3.73,
        "TYR-TYR": 4.17,
        "TYR-ALA": 3.36,
        "TYR-GLY": 3.01,
        "TYR-THR": 3.01,
        "TYR-SER": 2.78,
        "TYR-ASN": 2.76,
        "TYR-GLN": 2.97,
        "TYR-ASP": 2.76,
        "TYR-GLU": 2.79,
        "TYR-HIS": 3.52,
        "TYR-ARG": 3.16,
        "TYR-LYS": 2.60,
        "TYR-PRO": 3.19,
        "ALA-ALA": 2.72,
        "ALA-GLY": 2.31,
        "ALA-THR": 2.32,
        "ALA-SER": 2.01,
        "ALA-ASN": 1.84,
        "ALA-GLN": 1.89,
        "ALA-ASP": 1.70,
        "ALA-GLU": 1.51,
        "ALA-HIS": 2.41,
        "ALA-ARG": 1.83,
        "ALA-LYS": 1.31,
        "ALA-PRO": 2.03,
        "GLY-GLY": 2.24,
        "GLY-THR": 2.08,
        "GLY-SER": 1.82,
        "GLY-ASN": 1.74,
        "GLY-GLN": 1.66,
        "GLY-ASP": 1.59,
        "GLY-GLU": 1.22,
        "GLY-HIS": 2.15,
        "GLY-ARG": 1.72,
        "GLY-LYS": 1.15,
        "GLY-PRO": 1.87,
        "THR-THR": 2.12,
        "THR-SER": 1.96,
        "THR-ASN": 1.88,
        "THR-GLN": 1.90,
        "THR-ASP": 1.80,
        "THR-GLU": 1.74,
        "THR-HIS": 2.42,
        "THR-ARG": 1.90,
        "THR-LYS": 1.31,
        "THR-PRO": 1.90,
        "SER-SER": 1.67,
        "SER-ASN": 1.58,
        "SER-GLN": 1.49,
        "SER-ASP": 1.63,
        "SER-GLU": 1.48,
        "SER-HIS": 2.11,
        "SER-ARG": 1.62,
        "SER-LYS": 1.05,
        "SER-PRO": 1.57,
        "ASN-ASN": 1.68,
        "ASN-GLN": 1.71,
        "ASN-ASP": 1.68,
        "ASN-GLU": 1.51,
        "ASN-HIS": 2.08,
        "ASN-ARG": 1.64,
        "ASN-LYS": 1.21,
        "ASN-PRO": 1.53,
        "GLN-GLN": 1.54,
        "GLN-ASP": 1.46,
        "GLN-GLU": 1.42,
        "GLN-HIS": 1.98,
        "GLN-ARG": 1.80,
        "GLN-LYS": 1.29,
        "GLN-PRO": 1.73,
        "ASP-ASP": 1.21,
        "ASP-GLU": 1.02,
        "ASP-HIS": 2.32,
        "ASP-ARG": 2.29,
        "ASP-LYS": 1.68,
        "ASP-PRO": 1.33,
        "GLU-GLU": 0.91,
        "GLU-HIS": 2.15,
        "GLU-ARG": 2.27,
        "GLU-LYS": 1.80,
        "GLU-PRO": 1.26,
        "HIS-HIS": 3.05,
        "HIS-ARG": 2.16,
        "HIS-LYS": 1.35,
        "HIS-PRO": 2.25,
        "ARG-ARG": 1.55,
        "ARG-LYS": 0.59,
        "ARG-PRO": 1.70,
        "LYS-LYS": 0.12,
        "LYS-PRO": 0.97,
        "PRO-PRO": 1.75,
        "MET-CYS": 4.99,
        "PHE-CYS": 5.80,
        "ILE-CYS": 5.50,
        "LEU-CYS": 5.83,
        "VAL-CYS": 4.96,
        "TRP-CYS": 4.95,
        "TYR-CYS": 4.16,
        "ALA-CYS": 3.57,
        "GLY-CYS": 3.16,
        "THR-CYS": 3.11,
        "SER-CYS": 2.86,
        "ASN-CYS": 2.59,
        "GLN-CYS": 2.85,
        "ASP-CYS": 2.41,
        "GLU-CYS": 2.27,
        "HIS-CYS": 3.60,
        "ARG-CYS": 2.57,
        "LYS-CYS": 1.95,
        "PRO-CYS": 3.07,
        "PHE-MET": 6.56,
        "ILE-MET": 6.02,
        "LEU-MET": 6.41,
        "VAL-MET": 5.32,
        "TRP-MET": 5.55,
        "TYR-MET": 4.91,
        "ALA-MET": 3.94,
        "GLY-MET": 3.39,
        "THR-MET": 3.51,
        "SER-MET": 3.03,
        "ASN-MET": 2.95,
        "GLN-MET": 3.30,
        "ASP-MET": 2.57,
        "GLU-MET": 2.89,
        "HIS-MET": 3.98,
        "ARG-MET": 3.12,
        "LYS-MET": 2.48,
        "PRO-MET": 3.56,
        "ILE-PHE": 6.84,
        "LEU-PHE": 7.28,
        "VAL-PHE": 6.29,
        "TRP-PHE": 6.16,
        "TYR-PHE": 5.66,
        "ALA-PHE": 4.81,
        "GLY-PHE": 4.13,
        "THR-PHE": 4.28,
        "SER-PHE": 4.02,
        "ASN-PHE": 3.75,
        "GLN-PHE": 4.10,
        "ASP-PHE": 3.48,
        "GLU-PHE": 3.56,
        "HIS-PHE": 4.77,
        "ARG-PHE": 3.98,
        "LYS-PHE": 3.36,
        "PRO-PHE": 4.25,
        "LEU-ILE": 7.04,
        "VAL-ILE": 6.05,
        "TRP-ILE": 5.78,
        "TYR-ILE": 5.25,
        "ALA-ILE": 4.58,
        "GLY-ILE": 3.78,
        "THR-ILE": 4.03,
        "SER-ILE": 3.52,
        "ASN-ILE": 3.24,
        "GLN-ILE": 3.67,
        "ASP-ILE": 3.17,
        "GLU-ILE": 3.27,
        "HIS-ILE": 4.14,
        "ARG-ILE": 3.63,
        "LYS-ILE": 3.01,
        "PRO-ILE": 3.76,
        "VAL-LEU": 6.48,
        "TRP-LEU": 6.14,
        "TYR-LEU": 5.67,
        "ALA-LEU": 4.91,
        "GLY-LEU": 4.16,
        "THR-LEU": 4.34,
        "SER-LEU": 3.92,
        "ASN-LEU": 3.74,
        "GLN-LEU": 4.04,
        "ASP-LEU": 3.40,
        "GLU-LEU": 3.59,
        "HIS-LEU": 4.54,
        "ARG-LEU": 4.03,
        "LYS-LEU": 3.37,
        "PRO-LEU": 4.20,
        "TRP-VAL": 5.18,
        "TYR-VAL": 4.62,
        "ALA-VAL": 4.04,
        "GLY-VAL": 3.38,
        "THR-VAL": 3.46,
        "SER-VAL": 3.05,
        "ASN-VAL": 2.83,
        "GLN-VAL": 3.07,
        "ASP-VAL": 2.48,
        "GLU-VAL": 2.67,
        "HIS-VAL": 3.58,
        "ARG-VAL": 3.07,
        "LYS-VAL": 2.49,
        "PRO-VAL": 3.32,
        "TYR-TRP": 4.66,
        "ALA-TRP": 3.82,
        "GLY-TRP": 3.42,
        "THR-TRP": 3.22,
        "SER-TRP": 2.99,
        "ASN-TRP": 3.07,
        "GLN-TRP": 3.11,
        "ASP-TRP": 2.84,
        "GLU-TRP": 2.99,
        "HIS-TRP": 3.98,
        "ARG-TRP": 3.41,
        "LYS-TRP": 2.69,
        "PRO-TRP": 3.73,
        "ALA-TYR": 3.36,
        "GLY-TYR": 3.01,
        "THR-TYR": 3.01,
        "SER-TYR": 2.78,
        "ASN-TYR": 2.76,
        "GLN-TYR": 2.97,
        "ASP-TYR": 2.76,
        "GLU-TYR": 2.79,
        "HIS-TYR": 3.52,
        "ARG-TYR": 3.16,
        "LYS-TYR": 2.60,
        "PRO-TYR": 3.19,
        "GLY-ALA": 2.31,
        "THR-ALA": 2.32,
        "SER-ALA": 2.01,
        "ASN-ALA": 1.84,
        "GLN-ALA": 1.89,
        "ASP-ALA": 1.70,
        "GLU-ALA": 1.51,
        "HIS-ALA": 2.41,
        "ARG-ALA": 1.83,
        "LYS-ALA": 1.31,
        "PRO-ALA": 2.03,
        "THR-GLY": 2.08,
        "SER-GLY": 1.82,
        "ASN-GLY": 1.74,
        "GLN-GLY": 1.66,
        "ASP-GLY": 1.59,
        "GLU-GLY": 1.22,
        "HIS-GLY": 2.15,
        "ARG-GLY": 1.72,
        "LYS-GLY": 1.15,
        "PRO-GLY": 1.87,
        "SER-THR": 1.96,
        "ASN-THR": 1.88,
        "GLN-THR": 1.90,
        "ASP-THR": 1.80,
        "GLU-THR": 1.74,
        "HIS-THR": 2.42,
        "ARG-THR": 1.90,
        "LYS-THR": 1.31,
        "PRO-THR": 1.90,
        "ASN-SER": 1.58,
        "GLN-SER": 1.49,
        "ASP-SER": 1.63,
        "GLU-SER": 1.48,
        "HIS-SER": 2.11,
        "ARG-SER": 1.62,
        "LYS-SER": 1.05,
        "PRO-SER": 1.57,
        "GLN-ASN": 1.71,
        "ASP-ASN": 1.68,
        "GLU-ASN": 1.51,
        "HIS-ASN": 2.08,
        "ARG-ASN": 1.64,
        "LYS-ASN": 1.21,
        "PRO-ASN": 1.53,
        "ASP-GLN": 1.46,
        "GLU-GLN": 1.42,
        "HIS-GLN": 1.98,
        "ARG-GLN": 1.80,
        "LYS-GLN": 1.29,
        "PRO-GLN": 1.73,
        "GLU-ASP": 1.02,
        "HIS-ASP": 2.32,
        "ARG-ASP": 2.29,
        "LYS-ASP": 1.68,
        "PRO-ASP": 1.33,
        "HIS-GLU": 2.15,
        "ARG-GLU": 2.27,
        "LYS-GLU": 1.80,
        "PRO-GLU": 1.26,
        "ARG-HIS": 2.16,
        "LYS-HIS": 1.35,
        "PRO-HIS": 2.25,
        "LYS-ARG": 0.59,
        "PRO-ARG": 1.70,
        "PRO-LYS": 0.97,
    }

    updatedEnergyTable = {key: -value + 2.27 for key, value in energyTable.items()}

    if filename[-4:] != ".pdb":
        filename = filename + ".pdb"
    chains = []
    residues = []
    atoms = []
    interfaces = []
    with open(filename, "r") as fileName:
        currentChain = ""
        atomIndex = -1
        residueGlobalIndex = -1
        residueLocalIndex = -1
        chainIndex = -1
        for line in fileName:
            # parse one line from the pdb file
            typeName = line[0:4].strip()

            if typeName == "ATOM":
                index = int(line[6:11])
                atomType = line[12:16].strip()
                residueType = line[17:20].strip().upper()
                chainName = line[21]

                # check if it is a new chain
                if chainName != currentChain:
                    chainIndex += 1
                    chain = Chain(chainName, chainIndex)
                    currentChain = chainName
                    chains.append(chain)

                residueIndex = int(line[22:26])

                # check if it is a new residue
                if residueIndex != residueLocalIndex:
                    residueGlobalIndex += 1
                    residueLocalIndex = residueIndex
                    residue = Residue(residueGlobalIndex)
                    residue.chain = chainIndex
                    residues.append(residue)
                    chains[-1].residues.append(residueGlobalIndex)

                x = float(line[30:38])
                y = float(line[38:46])
                z = float(line[46:54])
                coord = Coord(x, y, z)
                atomIndex += 1
                atom = Atom(
                    atomIndex,
                    atomType,
                    residueType,
                    chainName,
                    chainIndex,
                    residueGlobalIndex,
                    coord,
                )
                atoms.append(atom)
                residues[-1].atoms.append(atomIndex)
                if atomType == "CA":
                    residues[-1].CA = atomIndex
                if atomType == "C1'":
                    residues[-1].isDNA = True
                    residues[-1].CA = atomIndex

    # calculate the center of mass for each chain
    for chain in chains:
        x = 0
        y = 0
        z = 0
        count = 0
        for residueIndex in chain.residues:
            residue = residues[residueIndex]
            for atomIndex in residue.atoms:
                chain.atoms.append(atomIndex)
                atom = atoms[atomIndex]
                x += atom.coord.x
                y += atom.coord.y
                z += atom.coord.z
                count += 1
        x /= count
        y /= count
        z /= count
        chain.COM = Coord(x, y, z)

    # calculate the interface
    interfaceIndex = -1
    for chain1 in chains:
        for chain2 in chains:
            if chain1.index < chain2.index:
                ca1 = []
                ca2 = []
                energy = []
                residuesPair = ()
                residues1 = ()
                residues2 = ()
                # loop all the atoms pair between chain1 and chain2
                for a1 in chain1.atoms:
                    for a2 in chain2.atoms:
                        atom1 = atoms[a1]
                        atom2 = atoms[a2]
                        dsquare = atom1.coord.distanceSquare(atom2.coord)
                        if dsquare <= cutoff * cutoff:
                            # ca1.append(residues[atom1.resideIndex].CA)
                            # ca2.append(residues[atom2.resideIndex].CA)
                            ca1.append(a1)
                            ca2.append(a2)
                            residuesPairName = (
                                atom1.chainName
                                + str(atom1.resideIndex)
                                + atom1.residueType
                                + atom2.chainName
                                + str(atom2.resideIndex)
                                + atom2.residueType
                            )
                            if residuesPairName not in residuesPair:
                                residuesPair += (residuesPairName,)
                                bond = atom1.residueType + "-" + atom2.residueType
                                if bond in updatedEnergyTable:
                                    energy.append(updatedEnergyTable[bond])
                            residues1 += (atom1.resideIndex,)
                            residues2 += (atom2.resideIndex,)
                if (
                    len(ca1) > 0
                    and len(residues1) >= thresholdResidue
                    and len(residues2) >= thresholdResidue
                ):
                    x, y, z = 0, 0, 0
                    for a in ca1:
                        CAIndex = residues[atoms[a].resideIndex].CA
                        x += atoms[CAIndex].coord.x
                        y += atoms[CAIndex].coord.y
                        z += atoms[CAIndex].coord.z
                    x /= len(ca1)
                    y /= len(ca1)
                    z /= len(ca1)
                    interfaceIndex += 1
                    interface = Interface(interfaceIndex, chain2.index)
                    interface.name = chain1.name + chain2.name
                    interface.chain = chain1.index
                    interface.coord = Coord(x, y, z)
                    interface.energy = sum(energy)
                    interface.atomsNum = len(ca1)
                    interface.residuesNum = len(residuesPair)
                    interface.size = len(residues1)
                    interfaces.append(interface)
                    chain1.interfaces.append(interfaceIndex)

                    x, y, z = 0, 0, 0
                    for a in ca2:
                        CAIndex = residues[atoms[a].resideIndex].CA
                        x += atoms[CAIndex].coord.x
                        y += atoms[CAIndex].coord.y
                        z += atoms[CAIndex].coord.z
                    x /= len(ca2)
                    y /= len(ca2)
                    z /= len(ca2)
                    interfaceIndex += 1
                    interface = Interface(interfaceIndex, chain1.index)
                    interface.name = chain2.name + chain1.name
                    interface.chain = chain2.index
                    interface.coord = Coord(x, y, z)
                    interface.energy = sum(energy)
                    interface.atomsNum = len(ca2)
                    interface.residuesNum = len(residuesPair)
                    interface.size = len(residues2)
                    interfaces.append(interface)
                    chain2.interfaces.append(interfaceIndex)
    # Move the interfaces to center of mass of the chain based on the stretch factor
    for chain in chains:
        for interfaceIndex in chain.interfaces:
            interface = interfaces[interfaceIndex]
            interface.coord.x = (
                interface.coord.x - chain.COM.x
            ) * stretchFactor + chain.COM.x
            interface.coord.y = (
                interface.coord.y - chain.COM.y
            ) * stretchFactor + chain.COM.y
            interface.coord.z = (
                interface.coord.z - chain.COM.z
            ) * stretchFactor + chain.COM.z

    # print the center of mass and interfaces for each chain
    for chain in chains:
        print(
            f"COM of chain {chain.name}: {chain.COM.x / 10.0:.3f}, {chain.COM.y / 10.0:.3f}, {chain.COM.z / 10.0:.3f}"
        )
        print(f"Interfaces of chain {chain.name}: ", end="")
        for interfaceIndex in chain.interfaces:
            print(interfaces[interfaceIndex].name, end=" ")
            print(
                f"partner chain: {chains[interfaces[interfaceIndex].partner].name}",
                end=" ",
            )
            print(
                f"{interfaces[interfaceIndex].coord.x / 10.0:.3f} {interfaces[interfaceIndex].coord.y / 10.0:.3f} {interfaces[interfaceIndex].coord.z / 10.0:.3f}",
                end=" ",
            )
            print(f"energy: {interfaces[interfaceIndex].energy:.3f}")
            print(f"residuesNum: {interfaces[interfaceIndex].residuesNum}")
        print()

    # write the pdb file with the interfaces added
    # Open a PDB file for writing
    with open("output.pdb", "w") as pdb_file:
        atom_serial = 1  # Initialize atom serial number
        residue_serial = 1  # Initialize residue serial number
        conect_records = []  # List to store CONECT records

        for chain in chains:
            com_serial = atom_serial  # Store the atom serial number for the COM

            # Write COM to PDB file
            pdb_file.write(
                f"ATOM  {atom_serial:5d} COM  COM {chain.name}{residue_serial:4d}    {chain.COM.x:8.3f}{chain.COM.y:8.3f}{chain.COM.z:8.3f}\n"
            )
            atom_serial += 1

            for interfaceIndex in chain.interfaces:
                interface = interfaces[interfaceIndex]

                # Write interface to PDB file
                pdb_file.write(
                    f"ATOM  {atom_serial:5d} INT  {interface.name.ljust(3)} {chain.name}{residue_serial:4d}    {interface.coord.x:8.3f}{interface.coord.y:8.3f}{interface.coord.z:8.3f}\n"
                )

                # Add bond between COM and interface
                conect_records.append(f"CONECT{com_serial:5d}{atom_serial:5d}\n")

                atom_serial += 1
            residue_serial += 1

        # Write all the CONECT records to the PDB file
        for conect in conect_records:
            pdb_file.write(conect)
    print("output.pdb has been generated.")

    print("nerdss input parameters:")

    # generate the input parameters for the nerdss simulation
    for chain in chains:
        print(f"mol {chain.name}:")
        print("  com 0.000 0.000 0.000")
        for i in chain.interfaces:
            interface = interfaces[i]
            print(
                f"  {interface.name} [{(interface.coord.x - chain.COM.x) / 10.0:.3f}, {(interface.coord.y - chain.COM.y) / 10.0:.3f}, {(interface.coord.z - chain.COM.z) / 10.0:.3f}]"
            )
            partner = None
            for partnerChain in chains:
                if partnerChain.index == interface.partner:
                    print(f"      partner {partnerChain.name}")
                    for partnerInterface in partnerChain.interfaces:
                        if interfaces[partnerInterface].partner == chain.index:
                            partner = interfaces[partnerInterface]
                            break
            print(f"      partner interface: {partner.name}")
            theta1, theta2, phi1, phi2, omega = calculateAngles(
                np.array([chain.COM.x, chain.COM.y, chain.COM.z]),
                np.array(
                    [
                        chains[interface.partner].COM.x,
                        chains[interface.partner].COM.y,
                        chains[interface.partner].COM.z,
                    ]
                ),
                np.array([interface.coord.x, interface.coord.y, interface.coord.z]),
                np.array([partner.coord.x, partner.coord.y, partner.coord.z]),
                np.array([0, 0, 1]),
                np.array([0, 0, 1]),
            )
            print(
                f"      theta1 {theta1:.3f} theta2 {theta2:.3f} phi1 {phi1:.3f} phi2 {phi2:.3f} omega {omega:.3f}"
            )
            print(
                f"      [{theta1:.3f}, {theta2:.3f}, {phi1:.3f}, {phi2:.3f}, {omega:.3f}]"
            )
            print("      n1 0.000 0.000 1.000")
            print("      n2 0.000 0.000 1.000")
            sigma = np.linalg.norm(
                np.array([interface.coord.x, interface.coord.y, interface.coord.z])
                - np.array([partner.coord.x, partner.coord.y, partner.coord.z])
            )
            print(f"       sigma {sigma / 10.0:.3f}")
            print(f"       energy {interface.energy:.3f}")
            print(f"       size(residuesNum) {interface.size}")
    # generate the input parameters for the nerdss simulation
    # Build the NERDSS molecule
    molecules = []
    reactions = []
    for chain in chains:
        molecules.append(Molecule(chain.name))
        molecules[-1].COM = chain.COM
        for interfaceIndex in chain.interfaces:
            interface = interfaces[interfaceIndex]
            molecules[-1].interfaces.append(Intf(interface.name, interface.coord))

    # Update the visual coord
    for molecule in molecules:
        for interface in molecule.interfaces:
            interface.positionsVisual = interface.positions
        molecule.COMVisual = molecule.COM

    # Shift the COM of molecule to [0,0,0]
    for molecule in molecules:
        for interface in molecule.interfaces:
            interface.positions -= [molecule.COM.x, molecule.COM.y, molecule.COM.z]
            interface.positions = [p / 10.0 for p in interface.positions]
        molecule.COM = Coord(0, 0, 0)

    # Calculate the radius of the molecule
    for molecule in molecules:
        for interface in molecule.interfaces:
            molecule.radius = max(molecule.radius, np.linalg.norm(interface.positions))

    # Calculate the translational diffusion coefficient and rotational diffusion coefficient
    for molecule in molecules:
        molecule.D = 226.95 / molecule.radius
        molecule.Dr = 170.22 / (molecule.radius**3)

    # Build the NERDSS reaction
    builtReactions = ()
    for chain in chains:
        mol1 = chain.name
        interfaceName1 = None
        mol2 = None
        interfaceName2 = None
        for i in chain.interfaces:
            interface = interfaces[i]
            interfaceName1 = interface.name
            partnerInterface = None
            for partnerChain in chains:
                if partnerChain.index == interface.partner:
                    mol2 = partnerChain.name
                    for j in partnerChain.interfaces:
                        partnerInterface = interfaces[j]
                        if partnerInterface.partner == chain.index:
                            interfaceName2 = partnerInterface.name
                            break
                    break
            reactionName = None
            if mol1 < mol2:
                reactionName = (
                    mol1
                    + "("
                    + interfaceName1
                    + ")"
                    + "-"
                    + mol2
                    + "("
                    + interfaceName2
                    + ")"
                )
            elif mol1 > mol2:
                reactionName = (
                    mol2
                    + "("
                    + interfaceName2
                    + ")"
                    + "-"
                    + mol1
                    + "("
                    + interfaceName1
                    + ")"
                )
            else:
                if interfaceName1 < interfaceName2:
                    reactionName = (
                        mol1
                        + "("
                        + interfaceName1
                        + ")"
                        + "-"
                        + mol2
                        + "("
                        + interfaceName2
                        + ")"
                    )
                else:
                    reactionName = (
                        mol2
                        + "("
                        + interfaceName2
                        + ")"
                        + "-"
                        + mol1
                        + "("
                        + interfaceName1
                        + ")"
                    )
            if reactionName not in builtReactions:
                builtReactions += (reactionName,)
                reaction = Reaction(
                    reactionName,
                    [
                        mol1 + "(" + interfaceName1 + ")",
                        mol2 + "(" + interfaceName2 + ")",
                    ],
                    mol1
                    + "("
                    + interfaceName1
                    + "!1)"
                    + "."
                    + mol2
                    + "("
                    + interfaceName2
                    + "!1)",
                )
                reaction.angles = calculateAngles(
                    np.array([chain.COM.x, chain.COM.y, chain.COM.z]),
                    np.array(
                        [
                            chains[interface.partner].COM.x,
                            chains[interface.partner].COM.y,
                            chains[interface.partner].COM.z,
                        ]
                    ),
                    np.array([interface.coord.x, interface.coord.y, interface.coord.z]),
                    np.array(
                        [
                            partnerInterface.coord.x,
                            partnerInterface.coord.y,
                            partnerInterface.coord.z,
                        ]
                    ),
                    np.array([0, 0, 1]),
                    np.array([0, 0, 1]),
                )
                reaction.norm1 = [0, 0, 1]
                reaction.norm2 = [0, 0, 1]
                reaction.sigma = np.linalg.norm(
                    np.array([interface.coord.x, interface.coord.y, interface.coord.z])
                    - np.array(
                        [
                            partnerInterface.coord.x,
                            partnerInterface.coord.y,
                            partnerInterface.coord.z,
                        ]
                    )
                )
                reaction.sigma = reaction.sigma / 10.0
                reaction.energy = interface.energy
                reaction.size = [interface.size, partnerInterface.size]
                reactions.append(reaction)
    # Write the NERDSS input file
    # get the current path
    folder = os.getcwd()
    # create the nerdss_inputs folder
    folder = os.path.join(folder, "nerdss_inputs")
    if not os.path.exists(folder):
        os.mkdir(folder)

    # Save the molecule file
    for mol in molecules:
        with open(os.path.join(folder, mol.name + ".mol"), "w") as mol_file:
            mol_file.write(f"Name = {mol.name}\n")
            mol_file.write("checkOverlap = true\n\n")
            mol_file.write(f"D = [{mol.D:.2f}, {mol.D:.2f}, {mol.D:.2f}]\n\n")
            mol_file.write(f"Dr = [{mol.Dr:.2f}, {mol.Dr:.2f}, {mol.Dr:.2f}]\n\n")
            mol_file.write("COM\t0.0000\t0.0000\t0.0000\n")
            for intf in mol.interfaces:
                mol_file.write(
                    f"{intf.name}\t{intf.positions[0]:.4f}\t{intf.positions[1]:.4f}\t{intf.positions[2]:.4f}\n"
                )
            mol_file.write("\n")
            mol_file.write(f"bonds = {len(mol.interfaces)}\n")
            for intf in mol.interfaces:
                mol_file.write(f"com {intf.name}\n")

    # Save the parameter file
    with open(os.path.join(folder, "parms.inp"), "w") as parm_file:
        parm_file.write("start parameters\n")
        parm_file.write("\tnItr = 1000000\n")
        parm_file.write("\ttimeStep = 0.1\n")
        parm_file.write("\ttimeWrite = 100\n")
        parm_file.write("\ttrajWrite = 100000\n")
        parm_file.write("\tpdbWrite = 1000\n")
        parm_file.write("\trestartWrite = 1000\n")
        parm_file.write("\tscaleMaxDisplace = 100.0\n")
        parm_file.write("\toverlapSepLimit = 1.0\n")
        parm_file.write("end parameters\n\n")
        parm_file.write("start boundaries\n")
        parm_file.write("\tWaterBox = [500.0, 500.0, 500.0]\n")
        parm_file.write("end boundaries\n\n")
        parm_file.write("start molecules\n")
        for mol in molecules:
            parm_file.write(f"\t{mol.name} : 100\n")
        parm_file.write("end molecules\n\n")
        parm_file.write("start reactions\n")
        for rect in reactions:
            parm_file.write(
                f"\t{rect.reactants[0]} + {rect.reactants[1]} <-> {rect.products}\n"
            )
            parm_file.write("\t\tonRate3Dka = 10.0\n")
            parm_file.write("\t\toffRatekb = 1.0\n")
            parm_file.write(f"\t\tsigma = {rect.sigma:.4f}\n")
            parm_file.write(
                f"\t\tnorm1 = [{rect.norm1[0]:.4f}, {rect.norm1[1]:.4f}, {rect.norm1[2]:.4f}]\n"
            )
            parm_file.write(
                f"\t\tnorm2 = [{rect.norm2[0]:.4f}, {rect.norm2[1]:.4f}, {rect.norm2[2]:.4f}]\n"
            )
            parm_file.write(
                f"\t\tassocAngles = [{rect.angles[0]:.4f}, {rect.angles[1]:.4f}, {rect.angles[2]:.4f}, {rect.angles[3]:.4f}, {rect.angles[4]:.4f}]\n"
            )
            parm_file.write(f"\t\texcludeVolumeBound = False\n")
            parm_file.write(
                f"\t\t#Contact Energy = {rect.energy:.4f}\tInterfaces' size (residues) = {rect.size[0]}, {rect.size[1]}\n\n"
            )
        parm_file.write("end reactions\n")


def calculateAngles(c1, c2, p1, p2, n1, n2):
    """
    Determine the angles of the reaction (theta1, theta2, phi1, phi2, omega) given the coordinates of the two Center of Mass (c1 and c2) and two reaction sites (p1 and p2), and two norm vectors (n1 and n2).

    Parameters
    ----------
    c1 : numpy.array
        Center of Mass vector for the first molecule.
    c2 : numpy.array
        Center of Mass vector for the second molecule.
    p1 : numpy.array
        Reaction site vector for the first molecule.
    p2 : numpy.array
        Reaction site vector for the second molecule.
    n1 : numpy.array
        Norm vector for the first molecule.
    n2 : numpy.array
        Norm vector for the second molecule.

    Returns
    -------
    tuple
        The tuple (theta1, theta2, phi1, phi2, omega), where theta1, theta2, phi1, phi2, omega are the angles in radians.
    """
    v1 = p1 - c1
    v2 = p2 - c2
    sigma1 = p1 - p2
    sigma2 = -sigma1

    theta1 = np.arccos(
        np.dot(v1, sigma1) / (np.linalg.norm(v1) * np.linalg.norm(sigma1))
    )
    theta2 = np.arccos(
        np.dot(v2, sigma2) / (np.linalg.norm(v2) * np.linalg.norm(sigma2))
    )

    t1 = np.cross(v1, sigma1)
    t2 = np.cross(v1, n1)
    norm_t1 = t1 / np.linalg.norm(t1)
    norm_t2 = t2 / np.linalg.norm(t2)
    phi1 = np.arccos(np.dot(norm_t1, norm_t2))

    # the sign of phi1 is determined by the direction of t2 relative to the right-hand rule of cross product of v1 and t1
    if np.dot(np.cross(v1, t1), t2) > 0:
        phi1 = -phi1

    t1 = np.cross(v2, sigma2)
    t2 = np.cross(v2, n2)
    norm_t1 = t1 / np.linalg.norm(t1)
    norm_t2 = t2 / np.linalg.norm(t2)
    phi2 = np.arccos(np.dot(norm_t1, norm_t2))

    # the sign of phi2 is determined by the direction of t2 relative to the right-hand rule of cross product of v2 and t1
    if np.dot(np.cross(v2, t1), t2) > 0:
        phi2 = -phi2

    if not np.isclose(np.linalg.norm(np.cross(v1, sigma1)), 0) and not np.isclose(
        np.linalg.norm(np.cross(v2, sigma2)), 0
    ):
        t1 = np.cross(sigma1, v1)
        t2 = np.cross(sigma1, v2)
    else:
        t1 = np.cross(sigma1, n1)
        t2 = np.cross(sigma1, n2)

    omega = np.arccos(np.dot(t1, t2) / (np.linalg.norm(t1) * np.linalg.norm(t2)))

    # the sign of omega is determined by the direction of t2 relative to the right-hand rule of cross product of sigma1 and t1
    if np.dot(np.cross(sigma1, t1), t2) > 0:
        omega = -omega

    return theta1, theta2, phi1, phi2, omega



from ..gen_platonic.distance import distance


def dode_vert_leg_reduce(COM: float, leg: float, sigma: float):
    """
    Reduces the length of a dodecahedron leg based on the center of mass (COM), leg vector, and sigma value.

    This function reduces the length of a dodecahedron leg based on the provided center of mass (COM), leg vector, and
    sigma value. The reduction is performed by calculating a ratio based on the sigma value and the distance between the
    center of mass and the leg vector. The leg vector is then scaled by this ratio and added to the center of mass to
    obtain the reduced leg vector.

    Args:
        COM (float): The center of mass (COM) of the dodecahedron vertex.
        leg (float): The leg vector of the dodecahedron vertex.
        sigma (float): The sigma value for the dodecahedron vertex.

    Returns:
        list: A list containing the three reduced leg vector coordinates.
    """
    
    red_len = sigma/2
    ratio = 1 - red_len/distance(COM, leg)
    leg_red = []
    for i in range(0, 3):
        leg_red.append(round((leg[i] - COM[i])*ratio + COM[i], 8))
    return leg_red





from ..gen_platonic.distance import distance


def cube_vert_leg_reduce(COM: float, leg: float, sigma: float):
    """Reduces the length of a cube vertex leg based on the center of mass (COM) and sigma value.

    This function reduces the length of a cube vertex leg based on the given center of mass (COM) and sigma value, using
    the `distance` function from the `gen_platonic` module to calculate the initial leg length, and then applying a
    reduction ratio based on the calculated length and the specified sigma value.

    Args:
        COM (float): The center of mass (COM) coordinate of the cube vertex, represented as a float value.
        leg (float): The original leg coordinate of the cube vertex, represented as a float value.
        sigma (float): The sigma value for the cube vertex, used to calculate the reduction ratio, represented as a float value.

    Returns:
        list: A list containing three floating-point values representing the reduced leg coordinates of the cube vertex,
        after applying the reduction ratio to each coordinate.


    Example:
        cube_vert_leg_reduce([0.5, 0.5, 0.5], [1.0, 1.0, 1.0], 0.2)
        # Reduces the length of the leg coordinate of a cube vertex with a center of mass (COM) of [0.5, 0.5, 0.5], an
        # original leg coordinate of [1.0, 1.0, 1.0], and a sigma value of 0.2.
        # Returns a list containing three floating-point values representing the reduced leg coordinates of the cube vertex
        # after applying the reduction ratio to each coordinate.
    """
        
    red_len = sigma/2
    ratio = 1 - red_len/distance(COM, leg)
    leg_red = []
    for i in range(0, 3):
        leg_red.append(round((leg[i] - COM[i])*ratio + COM[i], 8))
    return leg_red





import matplotlib.pyplot as plt


def get_time_dependence(filename, components):
    times = []
    counts = []
    components = sorted(components)  # Sort the components to ensure consistent matching

    current_time = None
    count_for_time = 0

    with open(filename, "r") as f:
        for line in f:
            if "Time (s):" in line:
                if current_time is not None:
                    times.append(current_time)
                    counts.append(count_for_time)
                current_time = float(line.split(":")[1].strip())
                count_for_time = 0
            else:
                count, comp = line.split("\t")
                comp_list = sorted(comp.strip().split(". ")[:])
                comp_list[-1] = comp_list[-1][:-1]
                if components == comp_list:
                    count_for_time += int(count)

        # Add the last time point
        if current_time is not None:
            times.append(current_time)
            counts.append(count_for_time)

    return times, counts


def plot_time_dependence(times, counts):
    plt.plot(times, counts)
    plt.xlabel("Time (s)")
    plt.ylabel("Copy Number")
    plt.title("Time Dependence of Complex Copy Numbers")
    plt.show()



# Form implementation generated from reading ui file 'nerdss_install.ui'
#
# Created by: PyQt6 UI code generator 6.4.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_NERDSSInstall(object):
    def setupUi(self, NERDSSInstall):
        NERDSSInstall.setObjectName("NERDSSInstall")
        NERDSSInstall.resize(663, 480)
        self.label = QtWidgets.QLabel(parent=NERDSSInstall)
        self.label.setGeometry(QtCore.QRect(200, 40, 261, 31))
        self.label.setObjectName("label")
        self.lineEditInstallPath = QtWidgets.QLineEdit(parent=NERDSSInstall)
        self.lineEditInstallPath.setGeometry(QtCore.QRect(80, 80, 391, 41))
        self.lineEditInstallPath.setObjectName("lineEditInstallPath")
        self.pushButtonInstallPath = QtWidgets.QPushButton(parent=NERDSSInstall)
        self.pushButtonInstallPath.setGeometry(QtCore.QRect(490, 80, 81, 41))
        self.pushButtonInstallPath.setObjectName("pushButtonInstallPath")
        self.pushButtonCancel = QtWidgets.QPushButton(parent=NERDSSInstall)
        self.pushButtonCancel.setGeometry(QtCore.QRect(170, 180, 80, 25))
        self.pushButtonCancel.setObjectName("pushButtonCancel")
        self.pushButtonInstall = QtWidgets.QPushButton(parent=NERDSSInstall)
        self.pushButtonInstall.setGeometry(QtCore.QRect(340, 180, 80, 25))
        self.pushButtonInstall.setObjectName("pushButtonInstall")
        self.label_2 = QtWidgets.QLabel(parent=NERDSSInstall)
        self.label_2.setGeometry(QtCore.QRect(80, 250, 511, 61))
        self.label_2.setObjectName("label_2")
        self.label_3 = QtWidgets.QLabel(parent=NERDSSInstall)
        self.label_3.setGeometry(QtCore.QRect(80, 300, 511, 16))
        self.label_3.setObjectName("label_3")

        self.retranslateUi(NERDSSInstall)
        QtCore.QMetaObject.connectSlotsByName(NERDSSInstall)

    def retranslateUi(self, NERDSSInstall):
        _translate = QtCore.QCoreApplication.translate
        NERDSSInstall.setWindowTitle(_translate("NERDSSInstall", "Install NERDSS"))
        self.label.setText(_translate("NERDSSInstall", "Select the Installation Path for NERDSS"))
        self.pushButtonInstallPath.setText(_translate("NERDSSInstall", "Select path"))
        self.pushButtonCancel.setText(_translate("NERDSSInstall", "Cancel"))
        self.pushButtonInstall.setText(_translate("NERDSSInstall", "Install"))
        self.label_2.setText(_translate("NERDSSInstall", "Input your password in the terminal if prompted. The installation might take a while."))
        self.label_3.setText(_translate("NERDSSInstall", "Please wait for the success box to pop up."))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    NERDSSInstall = QtWidgets.QDialog()
    ui = Ui_NERDSSInstall()
    ui.setupUi(NERDSSInstall)
    NERDSSInstall.show()
    sys.exit(app.exec())



from ..gen_platonic.angle_cal import angle_cal
from .dode_face_leg_reduce_coor_gen import dode_face_leg_reduce_coor_gen
from .dode_face_input_coord import dode_face_input_coord


def dode_face_write(radius: float, sigma: float):
    """Generate input file for dodecahedron face-centered simulation.

    This function takes a radius and a sigma value as input parameters,
    and generates an input file for a dodecahedron face-centered simulation
    using the provided parameters. The input file is written to a file named
    'parm.inp' and contains information about simulation parameters,
    boundaries, molecules, and reactions.

    Args:
        radius (float): Radius of the dodecahedron.
        sigma (float): Sigma value.

    Returns:
        parm.inp/cube.mol file: inputs for NERDSS
    """
    
    COM, lg1, lg2, lg3, lg4, lg5, n = dode_face_input_coord(radius, sigma)
    coord = dode_face_leg_reduce_coor_gen(radius, sigma)
    theta1, theta2, phi1, phi2, omega = angle_cal(
        coord[0][0], coord[0][3], coord[4][0], coord[4][1])

    f = open('parm.inp', 'w')
    f.write(' # Input file (dodecahedron face-centered)\n\n')
    f.write('start parameters\n')
    f.write('    nItr = 10000000 #iterations\n')
    f.write('    timeStep = 0.1\n')
    f.write('    timeWrite = 10000\n')
    f.write('    pdbWrite = 10000\n')
    f.write('    trajWrite = 10000\n')
    f.write('    restartWrite = 50000\n')
    f.write('    checkPoint = 1000000\n')
    f.write('    overlapSepLimit = 7.0\n')
    f.write('end parameters\n\n')
    f.write('start boundaries\n')
    f.write('    WaterBox = [500,500,500]\n')
    f.write('end boundaries\n\n')
    f.write('start molecules\n')
    f.write('    dode : 200\n')
    f.write('end molecules\n\n')
    f.write('start reactions\n')
    f.write('    dode(lg1) + dode(lg1) <-> dode(lg1!1).dode(lg1!1)\n')
    f.write('    onRate3Dka = 2\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    dode(lg2) + dode(lg2) <-> dode(lg2!1).dode(lg2!1)\n')
    f.write('    onRate3Dka = 2\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    dode(lg3) + dode(lg3) <-> dode(lg3!1).dode(lg3!1)\n')
    f.write('    onRate3Dka = 2\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    dode(lg4) + dode(lg4) <-> dode(lg4!1).dode(lg4!1)\n')
    f.write('    onRate3Dka = 2\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    dode(lg5) + dode(lg5) <-> dode(lg5!1).dode(lg5!1)\n')
    f.write('    onRate3Dka = 2\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    dode(lg1) + dode(lg2) <-> dode(lg1!1).dode(lg2!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    dode(lg1) + dode(lg3) <-> dode(lg1!1).dode(lg3!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    dode(lg1) + dode(lg4) <-> dode(lg1!1).dode(lg4!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    dode(lg1) + dode(lg5) <-> dode(lg1!1).dode(lg5!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    dode(lg2) + dode(lg3) <-> dode(lg2!1).dode(lg3!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    dode(lg2) + dode(lg4) <-> dode(lg2!1).dode(lg4!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    dode(lg2) + dode(lg5) <-> dode(lg2!1).dode(lg5!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    dode(lg3) + dode(lg4) <-> dode(lg3!1).dode(lg4!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    dode(lg3) + dode(lg5) <-> dode(lg3!1).dode(lg5!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('    dode(lg4) + dode(lg5) <-> dode(lg4!1).dode(lg5!1)\n')
    f.write('    onRate3Dka = 4\n')
    f.write('    offRatekb = 2\n')
    f.write('    norm1 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    norm2 = [' + str(n[0]) + ', ' +
            str(n[1]) + ', ' + str(n[2]) + ']\n')
    f.write('    sigma = ' + str(float(sigma)) + '\n')
    f.write('    assocAngles = [' + str(theta1) + ', ' + str(theta2) +
            ', ' + str(phi1) + ', ' + str(phi2) + ', ' + str(omega) + ']\n')
    f.write('    observeLabel = leg\n')
    f.write('    bindRadSameCom = 5.0\n')
    f.write('\n')
    f.write('end reactions\n')

    f = open('dode.mol', 'w')
    f.write('##\n')
    f.write('# Dodecahedron (face-centered) information file.\n')
    f.write('##\n\n')
    f.write('Name = dode\n')
    f.write('checkOverlap = true\n\n')
    f.write('# translational diffusion constants\n')
    f.write('D = [13.0, 13.0, 13.0]\n\n')
    f.write('# rotational diffusion constants\n')
    f.write('Dr = [0.03, 0.03, 0.03]\n\n')
    f.write('# Coordinates\n')
    f.write('COM   ' + str(round(COM[0], 8)) + '   ' +
            str(round(COM[1], 8)) + '   ' + str(round(COM[2], 8)) + '\n')
    f.write('lg1   ' + str(round(lg1[0], 8)) + '   ' +
            str(round(lg1[1], 8)) + '   ' + str(round(lg1[2], 8)) + '\n')
    f.write('lg2   ' + str(round(lg2[0], 8)) + '   ' +
            str(round(lg2[1], 8)) + '   ' + str(round(lg2[2], 8)) + '\n')
    f.write('lg3   ' + str(round(lg3[0], 8)) + '   ' +
            str(round(lg3[1], 8)) + '   ' + str(round(lg3[2], 8)) + '\n')
    f.write('lg4   ' + str(round(lg4[0], 8)) + '   ' +
            str(round(lg4[1], 8)) + '   ' + str(round(lg4[2], 8)) + '\n')
    f.write('lg5   ' + str(round(lg5[0], 8)) + '   ' +
            str(round(lg5[1], 8)) + '   ' + str(round(lg5[2], 8)) + '\n')
    f.write('\n')
    f.write('# bonds\n')
    f.write('bonds = 5\n')
    f.write('com lg1\n')
    f.write('com lg2\n')
    f.write('com lg3\n')
    f.write('com lg4\n')
    f.write('com lg5\n')
    f.write('\n')


# DODECAHEDEON VERTEX AS COM




from ..gen_platonic.distance import distance
from .dode_vert_input_coord import dode_vert_input_coord
import numpy as np

def dode_vert_norm_input(radius: float, sigma: float):
    """
    Calculates normalized input values for a dodecahedron vertex based on radius and sigma values.

    This function calculates the normalized center of mass (COM) and leg vectors for a dodecahedron vertex based on the
    provided radius and sigma values. The normalized COM and leg vectors are obtained by calling the dode_vert_input_coord
    function to calculate the original COM and leg vectors, and then performing various calculations to normalize their
    values. The normalized COM and leg vectors are stored in numpy arrays and returned.

    Args:
        radius (float): The radius of the dodecahedron.
        sigma (float): The sigma value for the dodecahedron vertex.

    Returns:
        numpy.ndarray: A numpy array representing the normalized center of mass (COM) vector.
        numpy.ndarray: A numpy array representing the normalized first leg (lg1) vector.
        numpy.ndarray: A numpy array representing the normalized second leg (lg2) vector.
        numpy.ndarray: A numpy array representing the normalized third leg (lg3) vector.
        numpy.ndarray: A numpy array representing the normalized normal (n) vector.
    """
    
    COM, lg1, lg2, lg3, n = dode_vert_input_coord(radius, sigma)
    length = distance(lg1, lg2)
    dis1 = ((-length/2)**2+(-((length/2)*(3**0.5))/3)**2)**0.5
    dis2 = distance(COM, lg1)
    height = (dis2**2-dis1**2)**0.5
    lg1_ = np.array([-length/2, -((length/2)*(3**0.5))/3, -height])
    lg2_ = np.array([length/2, -((length/2)*(3**0.5))/3, -height])
    lg3_ = np.array([0, ((length/2)*(3**0.5))/3*2, -height])
    COM_ = np.array([0, 0, 0])
    n_ = np.array([0, 0, 1])
    return COM_, lg1_, lg2_, lg3_, n_





from ..gen_platonic.mid_pt import mid_pt


def cube_face_COM_coord(a: float, b: float, c: float, d: float):
    """Calculates the center of mass (COM) coordinate for a cube face.

    This function calculates the COM coordinate for a cube face defined by four input points (a, b, c, d), where a, b, c,
    and d are the coordinates of the vertices of the cube face. The calculation is based on the mid-point coordinates
    of the input points, as well as the mid-point coordinates of the pairs of input points. The `mid_pt` function from
    the `..gen_platonic.mid_pt` module is used for the mid-point calculations.

    Args:
        a (float): The x-coordinate of the first vertex of the cube face.
        b (float): The x-coordinate of the second vertex of the cube face.
        c (float): The x-coordinate of the third vertex of the cube face.
        d (float): The x-coordinate of the fourth vertex of the cube face.

    Returns:
        Float: The x-coordinate of the calculated COM coordinate of the cube face.

    Example:
        >>> cube_face_COM_coord(0.0, 1.0, 1.0, 0.0)
        0.5
    """
    mid_a = mid_pt(a, b)
    mid_b = mid_pt(b, c)
    mid_c = mid_pt(c, d)
    mid_d = mid_pt(d, a)
    COM_a = mid_pt(mid_a, mid_c)
    COM_b = mid_pt(mid_b, mid_d)
    if COM_a == COM_b:
        return COM_a
    else:
        return COM_a





# Form implementation generated from reading ui file 'advanced_options_parse_pdb.ui'
#
# Created by: PyQt6 UI code generator 6.4.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_DialogParseParam(object):
    def setupUi(self, DialogParseParam):
        DialogParseParam.setObjectName("DialogParseParam")
        DialogParseParam.resize(640, 480)
        self.pushButtonApply = QtWidgets.QPushButton(parent=DialogParseParam)
        self.pushButtonApply.setGeometry(QtCore.QRect(230, 290, 80, 25))
        self.pushButtonApply.setObjectName("pushButtonApply")
        self.pushButtonCancel = QtWidgets.QPushButton(parent=DialogParseParam)
        self.pushButtonCancel.setGeometry(QtCore.QRect(330, 290, 80, 25))
        self.pushButtonCancel.setObjectName("pushButtonCancel")
        self.layoutWidget = QtWidgets.QWidget(parent=DialogParseParam)
        self.layoutWidget.setGeometry(QtCore.QRect(190, 90, 155, 161))
        self.layoutWidget.setObjectName("layoutWidget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.layoutWidget)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.label = QtWidgets.QLabel(parent=self.layoutWidget)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.label_2 = QtWidgets.QLabel(parent=self.layoutWidget)
        self.label_2.setObjectName("label_2")
        self.verticalLayout.addWidget(self.label_2)
        self.label_3 = QtWidgets.QLabel(parent=self.layoutWidget)
        self.label_3.setObjectName("label_3")
        self.verticalLayout.addWidget(self.label_3)
        self.layoutWidget1 = QtWidgets.QWidget(parent=DialogParseParam)
        self.layoutWidget1.setGeometry(QtCore.QRect(340, 80, 110, 171))
        self.layoutWidget1.setObjectName("layoutWidget1")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.layoutWidget1)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.lineEditCutoff = QtWidgets.QLineEdit(parent=self.layoutWidget1)
        self.lineEditCutoff.setObjectName("lineEditCutoff")
        self.verticalLayout_2.addWidget(self.lineEditCutoff)
        self.lineEditThreshold = QtWidgets.QLineEdit(parent=self.layoutWidget1)
        self.lineEditThreshold.setObjectName("lineEditThreshold")
        self.verticalLayout_2.addWidget(self.lineEditThreshold)
        self.lineEditStretch = QtWidgets.QLineEdit(parent=self.layoutWidget1)
        self.lineEditStretch.setObjectName("lineEditStretch")
        self.verticalLayout_2.addWidget(self.lineEditStretch)

        self.retranslateUi(DialogParseParam)
        QtCore.QMetaObject.connectSlotsByName(DialogParseParam)

    def retranslateUi(self, DialogParseParam):
        _translate = QtCore.QCoreApplication.translate
        DialogParseParam.setWindowTitle(_translate("DialogParseParam", "Parse PDB parameters"))
        self.pushButtonApply.setText(_translate("DialogParseParam", "Apply"))
        self.pushButtonCancel.setText(_translate("DialogParseParam", "Cancle"))
        self.label.setText(_translate("DialogParseParam", "Cutoff distance"))
        self.label_2.setText(_translate("DialogParseParam", "Threshold for residues"))
        self.label_3.setText(_translate("DialogParseParam", "Stretch factor"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    DialogParseParam = QtWidgets.QDialog()
    ui = Ui_DialogParseParam()
    ui.setupUi(DialogParseParam)
    DialogParseParam.show()
    sys.exit(app.exec())



from .dode_vert_leg_reduce_coor_gen import dode_vert_leg_reduce_coor_gen
import numpy as np

def dode_vert_input_coord(radius: float, sigma: float):
    """Generates input coordinates for a dodecahedron vertex.

    This function generates input coordinates for a dodecahedron vertex, given the radius and sigma values. The input
    coordinates are calculated based on the radius and sigma values using the dode_vert_leg_reduce_coor_gen function
    from the .dode_vert_leg_reduce_coor_gen module.

    Args:
        radius (float): The radius of the dodecahedron vertex.
        sigma (float): The sigma value for the dodecahedron vertex.

    Returns:
        tuple: A tuple containing the following five numpy arrays:
            - COM: The center of mass (COM) of the dodecahedron vertex.
            - lg1: The first leg vector of the dodecahedron vertex.
            - lg2: The second leg vector of the dodecahedron vertex.
            - lg3: The third leg vector of the dodecahedron vertex.
            - n: The normalized vector of the dodecahedron vertex.
    """
    
    coor = dode_vert_leg_reduce_coor_gen(radius, sigma)
    coor_ = np.array(coor[0])
    COM = np.around(coor_[0] - coor_[0], 12)
    lg1 = np.around(coor_[1] - coor_[0], 12)
    lg2 = np.around(coor_[2] - coor_[0], 12)
    lg3 = np.around(coor_[3] - coor_[0], 12)
    n = np.around(coor_[0]/np.linalg.norm(coor_[0]), 12)
    return COM, lg1, lg2, lg3, n





import numpy as np
import os
import matplotlib.pyplot as plt
from ...file_managment.save_vars_to_file import save_vars_to_file

def multi_hist_complex_count(FileName: str = 'histogram_complexes_time.dat', FileNum: int = 1, InitialTime: float = 0, FinalTime: float = 1E10,
               SpeciesList: list = ['all'], BinNums: int = 10, ExcludeSize: int = 0, ShowFig: bool = True, SaveFig: bool = False, SaveVars: bool = False) -> tuple:
    """
    Generate histogram of the size of target species for a multiple species system for the given species 
    and targets from the histogram_complexes_time.dat in the input file within the specified time range.

    Args:
        FileName (str): The name of the input file. Default is 'histogram_complexes_time.dat'
        FileNum (int): The number of files to read. Default is 1.
        InitialTime (float): The start time of the time range (inclusive). Default is 0.
        FinalTime (float): The end time of the time range (exclusive). Default is 1E10.
        SpeciesList (list): The list of targets for whose number will be counted as the size of the complex. Default is all species.
        BinNums (int, optional): The number of bins in the histogram. Default is 10.
        ExcludeSize (int, optional): The minimum value required to include a data point in the histogram. Default is 0.
        ShowFig (bool, optional): Whether to display the generated figures. Default is True.
        SaveFig (bool, optional): Whether to save the generated figures. Default is False.
        SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.

    Returns:
        the np.arrays of normalized counts, means, and standard deviations for the histograms.
    """

    FileName_head, FileName_tail = os.path.splitext(FileName)

    # Create dictionary to store count of one size
    sizeDictList = []
     
    for i in range(1, FileNum+1):
        # check if there is a folder named '1' in the current folder, if there is, set the temp_FileName to ./i/FileName, then check if ./i/FileName exited. if not exited, throw out a file not exit error.
        # else if FileNum is 1, check if there is file named FileName in the current folder, if there is, set the temp_FileName to FileName; if FileNum > 1, check if there is afile name FileName_head + '_' + str(i) + '.' + FileName_tail, if it is, set as the temp_FileName
        # else throw a file not found error
        temp_FileName = ""
        if os.path.isdir(str(i)):
            temp_FileName = os.path.join(str(i), FileName)
            if not os.path.isfile(temp_FileName):
                raise FileNotFoundError(f"File {temp_FileName} does not exist.")
        elif FileNum == 1:
            if os.path.isfile(FileName):
                temp_FileName = FileName
            else:
                raise FileNotFoundError(f"File {FileName} does not exist.")
        else:
            temp_FileName = f"{FileName_head}_{i}{FileName_tail}"
            if not os.path.isfile(temp_FileName):
                raise FileNotFoundError(f"File {temp_FileName} does not exist.")

        sizeDict = {}
        # Open file and read line-by-line
        with open(temp_FileName, 'r') as file:
            found_InitialTime = False

            for line in file:
                # Check if line contains time information
                if line.startswith('Time (s):'):
                    time = float(line.split()[-1])   # Extract time value from line
                    if time < InitialTime:
                        continue
                    elif time >= FinalTime:
                        break
                    else:
                        found_InitialTime = True

                else:   # Line contains complex count information
                    if found_InitialTime:
                        # Extract count and complex information
                        count, complexInfoStr = line.strip().split('\t')
                        count = int(count)
                        # Split complex information into individual items
                        complexItems = complexInfoStr.rstrip('.').split('. ')
                        size = 0
                        for item in complexItems:
                            # Extract species name and count from complex information
                            speciesName, numStr = item.split(': ')
                            num = int(numStr)
                            if SpeciesList[0] == 'all' and len(SpeciesList) == 1:
                                size += num
                            else:
                                if speciesName in SpeciesList:
                                    size += num
                        if size > ExcludeSize:
                            if size not in sizeDict:
                                sizeDict[size] = count
                            else:
                                sizeDict[size] += count
        sizeDictList.append(sizeDict)
        
    # Combine counts of all sizes from all dictionaries
    combined_counts = {}
    for d in sizeDictList:
        for size, count in d.items():
            if size in combined_counts:
                combined_counts[size].append(count)
            else:
                combined_counts[size] = [count]

    # Calculate mean and standard deviation for each size
    mean_std = {}
    for size, counts in combined_counts.items():
        mean = np.mean(counts)
        std = np.std(counts)
        mean_std[size] = (mean, std)

    mean_std = dict(sorted(mean_std.items()))
    size_array = np.array(list(mean_std.keys()))
    mean_array = np.array([value[0] for value in mean_std.values()])
    std_array = np.array([value[1] for value in mean_std.values()])
    
    #output variables
    if SaveVars:
        save_vars_to_file({"cmplx_sizes":size_array, "cmplx_count":mean_array, "std":std_array})

    plt.hist(size_array, bins=BinNums, weights=mean_array, density=True, histtype='bar', alpha=0.75)

    plt.xlabel('Size of complex')
    plt.ylabel('Frequency')
    if SaveFig:
        plt.SaveFig('histgram_size_of_complex.png', dpi=300)
    if ShowFig:
        plt.show()
    return size_array, mean_array, std_array


import numpy as np


def xyz_to_sphere_coordinates(position): # translate x-y-z coords to spherical coords
    """
    Translate x-y-z coordinates to spherical coordinates.
    Args:
        position (list): x-y-z coordinates of a point
    Returns:
        spherecoordinates (list): theta-phi-r coordinates of the point
    """
    x = position[0]
    y = position[1]
    z = position[2]
    r = np.sqrt(x**2 + y**2 + z**2)
    theta = np.arccos(z/r)
    phi = np.arccos( x/r/np.sin(theta) )
    if y < 0 :
        phi = 2.0*np.pi - phi
    spherecoordinates = [theta, phi, r]
    return spherecoordinates




import math

def distance_bw_atoms_in_chains(split_position,split_resi_position,i,j, max_bound_length: float = 0.3):
    """Determines which atoms are bounded, and adds their position to a list

    Args:
        split_position (list): position of each atom in each chain
        split_resi_position (list): position of each residiual of each atom in each chain
        i (int): Index of the first chain
        j (int): Index of the second chain
        max_bound_length(float): atoms that are less than this length apart are seen as bound

    Returns:
        list: list of the positions of each atoms that are bounded
        int: count of how many atoms that are bounded b/w these chains
    """
    
    
    
    inner_reaction_resi_position = []
    count = 0 #if these 2 chains have any atoms binding b/w them
    
    
    #for each atom in chain 1 (i)
    for m,atom_coords_ch1 in enumerate(split_position[i]):
                
        #for each atom in chain 2 (j)
        for n,atom_coords_ch2 in enumerate(split_position[j]):
                    
            #calculate the distance between the two
            distance = math.sqrt((atom_coords_ch1[0]-atom_coords_ch2[0])**2 + (atom_coords_ch1[1]-atom_coords_ch2[1])**2 + (atom_coords_ch1[2]-atom_coords_ch2[2])**2)

            if distance <= max_bound_length:
                inner_reaction_resi_position.append([split_resi_position[i][m], split_resi_position[j][n]])
                count += 1
    return inner_reaction_resi_position,count





def real_PDB_chain_int_simple(unique_chain, split_position, split_resi_count, split_atom_count, split_resi_type, split_atom_type, split_resi_position, max_length_bound: float = 0.3):
    """
    This function takes a complex protein structure and determines which chains and residues are interacting
    with each other based on the distance between atoms. The output is a tuple that includes the following lists:

    Args:
        unique_chain (list): Unique chains within the protein structure
        split_position (list of list): Each sublist contains the positions of atoms in a specific chain
        split_resi_count (list of list): Each sublist contains the residue count of atoms in a specific chain
        split_atom_count (list of list): Each sublist contains the atom count in a specific chain
        split_resi_type (list of list): aEach sublist contains the residue type of atoms in a specific chain
        split_atom_type (list of list): Each sublist contains the atom type of atoms in a specific chain
        split_resi_position (list of list): Each sublist contains the residue position of atoms in a specific chain
        max_bound_length(float): atoms that are less than this length apart are seen as bound
    
    Returns:
        reaction_chain (list of list): Each sublist contains the chain IDs of two chains that are interacting with each other
        reaction_resi_position (list of list of list): Each sub-sublist contains pairs of residue positions that are interacting

    """
    
    # list of lists (each sublist will include two letters indicating these two chains have interaction)
    reaction_chain = []
    
    # list of lists of lists (Each sublist contains the residue position of atoms in a specific chain)
    reaction_resi_position = []
    
    #for each unique chain
    for i in range(len(unique_chain) - 1):
        
        #for each unique chain, not this one!
        for j in range(i+1, len(unique_chain)):
            
            #calculates distance between each atom in the chain, creating a list of their positions and overall number
            inner_reaction_resi_position,count = distance_bw_atoms_in_chains(split_position,split_resi_position,i,j, max_length_bound)
            
            if count > 0:
                reaction_chain.append([unique_chain[i], unique_chain[j]])
                reaction_resi_position.append(inner_reaction_resi_position)
    return reaction_chain, reaction_resi_position





from .tetr_vert_COM_leg_gen import tetr_vert_COM_leg_gen
from .tetr_vert_leg_reduce import tetr_vert_leg_reduce

def tetr_vert_leg_reduce_coor_gen(radius: float, sigma: float):
    """Generate the reduced leg coordinates of a regular tetrahedron given the radius and sigma.

    Args:
        radius (float): The radius of the circumsphere of the tetrahedron.
        sigma (float): The scaling factor for reducing the length of the legs.

    Returns:
        list: A list of lists containing the coordinates of the center of mass (COM) and the reduced leg vectors
        of the tetrahedron, for each of the four vertices.
    """
    # Generating all the coords of COM and legs when sigma exists
    COM_leg_list = tetr_vert_COM_leg_gen(radius)
    COM_leg_red_list = []
    for elements in COM_leg_list:
        temp_list = []
        temp_list.append(elements[0])
        i = 1
        while i <= 3:
            temp_list.append(tetr_vert_leg_reduce(
                elements[0], elements[i], sigma))
            i += 1
        COM_leg_red_list.append(temp_list)
    return COM_leg_red_list





def create_complex_list(bond_list):
    """Finds protein complexes based on a list of protein-protein bond pairs.

    Args:
        bond_lst (list): A list of protein-protein bond pairs in the format ([protein_num_1, protein_num_2]_.

    Returns:
        list: A list of protein complex lists, where each complex list contains protein numbers that are part of
        the same complex.
    """
    
    complex_lst = []
    
    #if the bond list is not empty, keep going through it
    while bond_list != []:
        temp_complex_lst = bond_list[0]

        #go through each the temp protein complex list. For each protein, check through the enitre bond_list to see if it is bonded with anything. If yes, add it and remove it from main list
        for protein in temp_complex_lst:
            for bond in bond_list:
                if protein == bond[0]:
                    temp_complex_lst.append(bond[1])
                    bond_list.remove(bond)
                elif protein == bond[1]:
                    temp_complex_lst.append(bond[0])
                    bond_list.remove(bond)

        
        temp_complex_lst.sort()
        complex_lst.append(list(set(temp_complex_lst)))


    
    
    return complex_lst


import numpy as np


def read_cluster_lifetime(FileName: str, SpeciesName: str, InitialTime: float, FinalTime: float):
    """Reads and extracts the lifetimes of clusters of a specific species from a transition_matrix_time.dat file.

    Args:
        FileName (str): The name of the file to read the cluster lifetimes from.
        SpeciesName (str): The name of the species to extract cluster lifetimes for.
        InitialTime (float): The initial time at which to extract cluster lifetimes.
        FinalTime (float): The final time at which to extract cluster lifetimes.

    Returns:
        A tuple containing three elements:
        - A list of numpy arrays representing the lifetimes of the species clusters
        at the initial time.
        - A list of numpy arrays representing the lifetimes of the species clusters
        at the final time.
        - A list of integers representing the sizes of the species clusters.
    """
    size_list = []
    ti_lifetime = []
    tf_lifetime = []
    
    ##This file initiall had everything in the with open() meaning it would go through each line, and check it. Now it will check only the start, than only will check lines. This gave it 
    ##A slight boost in speed for higher timestep data

    #read the file (slowest part I promise)
    with open(FileName, 'r') as file:
        lines = file.readlines()

    #determine distance between times
    for index,line in enumerate(lines[1:]):
        
        #distance b/w time line and the correct species name line
        if line[0:4] == SpeciesName:
            speciesDistance = index + 1

        #distance between 2 different time slimes
        if line[0:5] == 'time:': 
            distance = index + 1
            break     

        
    #for every time line
    for index,line in enumerate(lines[::distance]):
        
        #checks if it equal to the initial time
        if float(line.split(' ')[1]) == InitialTime:
            index_real = (distance*index) + 1
            index_start_read = index_real + speciesDistance   
            for data in lines[index_start_read:]:
                
                #if time is reached
                if data[0:5] == 'time:':
                    break

                #cluster size line
                if data[0:20] == 'size of the cluster:':
                    size_list.append(int(data.split(':')[1].strip('\n')))
                        
                #the lifetimes of that size??
                else:
                    str_list = data.strip('\n').strip(' ').split(' ')
                    temp = []
                    for i in str_list:
                        if i != '':
                            temp.append(float(i))
                    ti_lifetime.append(np.array(temp))

        
        #checks if it equal to the final time
        if float(line.split(' ')[1]) == FinalTime:
            index_real = (distance*index) + 1
            index_start_read = index_real + speciesDistance   

            for data in lines[index_start_read:]:
                
                #if time is reached
                if data[0:5] == 'time:':
                    break
                
                #the lifetimes of that size??
                if data[0:20] != 'size of the cluster:':
                    str_list = data.strip('\n').strip(' ').split(' ')
                    temp = np.array([])
                    for i in str_list:
                        if i != '':
                            temp = np.append(temp, float(i))
                    tf_lifetime.append(temp)
    
    return ti_lifetime, tf_lifetime, size_list





def mid_pt(a: float, b: float):
    """Compute the mid-point between two coordinates in 3-dimensional space.

    Parameters:
        a (float): The first coordinate in the form of [x, y, z].
        b (float): The second coordinate in the form of [x, y, z].

    Returns:
        List[float]: The mid-point coordinates in the form of [x, y, z].


    Examples
        >>> a = [0.0, 0.0, 0.0]
        >>> b = [2.0, 4.0, 6.0]
        >>> mid_pt(a, b)
        [1.0, 2.0, 3.0]

    Notes
        This function calculates the mid-point between two coordinates in 3-dimensional space
        by taking the average of the corresponding x, y, and z coordinates of the two points.
        The result is rounded to 15 decimal places using the `round()` function with `n` set to 15,
        which is the value of `n` used in the function implementation.
    """
    
    # this is a seperate function for calculating mid point of two coords
    n = 15
    return [round((a[0]+b[0])/2, n), round((a[1]+b[1])/2, n), round((a[2]+b[2])/2, n)]





def filter_complexes(complex_lst,num_name_dict,num_dict):
    """Generates a list of every protein in a complex of the correct size

    Args:
        complex_lst (List[List[str]]): a list of complexes, where each complex is represented as a list of protein names
        num_name_dict (dictionary): a dictionary that holds the protein type for each protein number
        num_dict (dictionary): a dictionary that holds the wanted amount of each protein type in the complex.

    Returns:
        list: a list of each protein number that is in a complex of the correct size
    """

    complex_filtered = []
    #run through every protein complex, and determine it's length.
    for complex in complex_lst:
        
        #create dictionary to hold the values
        temp_complex_num = dict.fromkeys(num_dict,0)
        
        #count up each protein type in the complex
        for protein in complex:
            if num_name_dict[str(protein)] in temp_complex_num.keys():
                temp_complex_num[num_name_dict[str(protein)]] += 1

        #if it has the corret number of proteins, add it to the new list
        if temp_complex_num == num_dict:
            for protein in complex:
                complex_filtered.append(protein)


    return complex_filtered

   


import os
import numpy as np


def add_empty_lists(input_list: list):
    """Will take in a list. If it has no value, it will add 'Empty List'. If it has sub lists, it will open them and do the same thing. (recursive)

    Args:
        input_list (list): list to be searched through
    """

    #check if list is empty
    if len(input_list) <= 0:
        input_list.append("Empty List")

    #check through all indexes to see if there are any lists. If there are run this function on them, else do nothing.
    input_list = [add_empty_lists(list(vars)) if (isinstance(vars,list) or isinstance(vars,np.ndarray)) else vars for vars in input_list]

    """ ^^^ equals what is below.
    for index,vars in enumerate(input_list):
        if isinstance(vars,list) or isinstance(vars,np.ndarray):
            input_list[index] = add_empty_lists(list(vars))"""

    return input_list



def save_vars_to_file(var_dict: dict):
    """This function takes a dictionary of variables and saves them to a file in the correct format.

        Parameters
            var_dict: A dictionary of variables to be saved.
        
        Explanation
            If the variable is a number or a string, it saves it to a .txt file. If the variable is a list, 
            it saves it to a .csv file. If the list is two-dimensional or deeper, it will save each sublist on a different line.
    """


    #create folder if necessary 
    file_start = "" #the starting part of the file

    #if there is a folder needed
    if len(var_dict) > 1:
        file_start = "vars/"

        #if the folder needs to be made
        if not os.path.exists("vars"):
            os.mkdir(path="vars")


    for key,value in var_dict.items():

        type = "" # what kind of file is being made
        file_end = "" # the ending part of the file
    
        #determine variable type
        if isinstance(value,int) or isinstance(value,float): #number
            file_end = f"{key}_number.txt"
            type = "txt"
        
        elif isinstance(value,str): #string
            file_end = f"{key}_string.txt"
            type = "txt"    
        
        elif isinstance(value,list) or isinstance(value,np.ndarray): #list
            file_end = f"{key}_list.csv"
            value = add_empty_lists(list(value)) #adds 'empty list' to lists that are empty 
            
            if isinstance(value[0],list) or isinstance(value[0],np.ndarray): #2dim+ deep list
                type = "2list"
            else: #1dim list
                type = "1list"
  
        else:
            file_end = f"{key}_na.txt"
            type = "txt"
        
        file_name = file_start + file_end



        #Create file name and open file 
        with open(file_name, mode = "w") as file:

            #if it is just basic text
            if type == "txt":
                file.write(str(value))
            
            #if it is a 1dim list
            elif type == "1list":
                file.write(str(value[0]))
                
                if len(value) > 1:
                    for var in value[1:]:
                        file.write(f",{var}")
            
            #if it is a 2dem list
            elif type == "2list":

                #run through each sub list
                for index,list1 in enumerate(value):
                    
                    #add \n, unless it is the first sublist
                    if index != 0:
                        file.write("\n")
                    
                    #write in entire sublist
                    file.write(f"{list1[0]}")
                    
                    if len(list1) > 1:
                        for list2 in list1[1:]:
                            file.write(f",{list2}")




            

        







import numpy as np


def calculate_gradient(centers,xyzR): 
    """calculate the gradient of the gag positions for the optimization of the gag positions on the sphere surface
         
    Args:
        centers (18*3 list of floats): the x, y, z coordinates of the 18 gag centers
        xyzR (list of floats): the x, y, z coordinates of the center of the gag sphere and the radius of the gagsphere

    Returns:
        gradient (list of floats): the gradient of the gag positions
    """

    x0 = xyzR[0]
    y0 = xyzR[1]
    z0 = xyzR[2]
    r0 = xyzR[3]
    dsdx = 0
    dsdy = 0
    dsdz = 0
    dsdr = 0
    for i in range(0,centers.shape[0]):
        xi = centers[i,0]
        yi = centers[i,1]
        zi = centers[i,2]
        ri = np.sqrt( (xi-x0)**2 + (yi-y0)**2 + (zi-z0)**2 )
        dsdx = dsdx + (-2.0/ri)*(ri-r0)*(xi-x0)
        dsdy = dsdy + (-2.0/ri)*(ri-r0)*(yi-y0)
        dsdz = dsdz + (-2.0/ri)*(ri-r0)*(zi-z0)
        dsdr = dsdr + (-2.0)*(ri-r0)
    gradient = np.array([dsdx, dsdy, dsdz, dsdr])
    return gradient




import pandas as pd
import os
from .xyz_to_csv import xyz_to_csv


def xyz_to_df(FileName: str, LitNum: int, SaveCsv: bool = True):
    """Converts a .xyz file to a pandas.DataFrame for a specific or entire time frame.

    Args:
        FileName (str): The path to the input .xyz file, usually named 'trajectory.xyz'.
        LitNum (int): The number of iterations to examine. If -1, the entire iteration will be extracted.
        SaveCsv (bool): Whether to save the corresponding .csv file (default: True).

    Returns:
        A pandas.DataFrame containing 5 columns: number of iteration, species name, x, y, and z coordinates.

    Description:
        This function enables users to convert the output .xyz file by NERDSS simulation into a pandas.DataFrame of a specific or entire time frame. The generated DataFrame will contain 5 columns: number of iteration, species name, x, y, and z coordinates.

    Sample:
        df = xyz_to_df('/Users/UserName/Documents/trajectory.xyz', 100000000, True) # Extracts iteration 100000000 and saves the corresponding .csv file
        df = xyz_to_df('/Users/UserName/Documents/trajectory.xyz', -1, False) # Extracts the entire iteration and deletes the generated .csv file
    """
    xyz_to_csv(FileName, LitNum)
    if LitNum != -1:
        write_file_name = 'trajectory_' + str(LitNum) + '.csv'
    else:
        write_file_name = 'trajectory_full.csv'
    df = pd.read_csv(write_file_name)
    if not SaveCsv:
        os.remove(write_file_name)
    return df





from .icos_vert_coord import icos_vert_coord
from .icos_vert_COM_leg import icos_vert_COM_leg

def icos_vert_COM_leg_gen(radius: float):
    """Generate a list of center of mass (COM) and legs coordinates for an icosahedron.

    The function calculates the center of mass and legs coordinates for an icosahedron
    with the given radius, using the `icos_vert_coord` and `icos_vert_COM_leg` functions.

    Args:
        radius (float): The radius of the icosahedron.

    Returns:
        list: A list of center of mass and legs coordinates for the icosahedron.
    """
    coord = icos_vert_coord(radius)
    COM_leg_list = []
    COM_leg_list.append(icos_vert_COM_leg(
        coord[0], coord[2], coord[8], coord[4], coord[6], coord[10]))
    COM_leg_list.append(icos_vert_COM_leg(
        coord[1], coord[4], coord[6], coord[11], coord[3], coord[9]))
    COM_leg_list.append(icos_vert_COM_leg(
        coord[2], coord[0], coord[10], coord[7], coord[5], coord[8]))
    COM_leg_list.append(icos_vert_COM_leg(
        coord[3], coord[1], coord[11], coord[7], coord[5], coord[9]))
    COM_leg_list.append(icos_vert_COM_leg(
        coord[4], coord[0], coord[6], coord[1], coord[9], coord[8]))
    COM_leg_list.append(icos_vert_COM_leg(
        coord[5], coord[2], coord[8], coord[7], coord[3], coord[9]))
    COM_leg_list.append(icos_vert_COM_leg(
        coord[6], coord[0], coord[10], coord[11], coord[1], coord[4]))
    COM_leg_list.append(icos_vert_COM_leg(
        coord[7], coord[3], coord[11], coord[10], coord[2], coord[5]))
    COM_leg_list.append(icos_vert_COM_leg(
        coord[8], coord[0], coord[2], coord[5], coord[9], coord[4]))
    COM_leg_list.append(icos_vert_COM_leg(
        coord[9], coord[8], coord[4], coord[1], coord[3], coord[5]))
    COM_leg_list.append(icos_vert_COM_leg(
        coord[10], coord[0], coord[2], coord[7], coord[11], coord[6]))
    COM_leg_list.append(icos_vert_COM_leg(
        coord[11], coord[10], coord[7], coord[3], coord[1], coord[6]))
    return COM_leg_list





from .tetr_face_write import tetr_face_write


def tetr_face(radius: float, sigma: float):
    """Draws a tetrahedron face with the given radius and sigma.

    Args:
        radius (float): The radius of the tetrahedron.
        sigma (float): The sigma value for drawing the tetrahedron face.

    Returns:
        parm.inp/icos.mol: input files for NERDSS

    Examples:
        >>> tetr_face(1.0, 0.5)
        File writing complete!

    Note:
        This function relies on the 'tetr_face_write' function from the '.tetr_face_write' module.
        The 'tetr_face_write' function is responsible for writing the tetrahedron face to a file.
    """
    tetr_face_write(radius, sigma)
    print('File writing complete!')
    return 0







def hist_temp(hist: list, InitialTime: float, FinalTime: float):
    """Determines the average amount of each complex type/size over this timestep

    Args:
        FileName (list): histogram data that has been read in
        InitialTime (float): The starting time. Must not be smaller / larger then times in file.
        FinalTime (float): The ending time. Must not be smaller / larger then times in file.

    Returns:
        list: a list of each complex type/size
        list: a list of the average amount of each complex type/size
    """
    
    #reads the file
   
    plot_count = [] #total count of that certain size
    plot_conv = [] #list of each complex size
    tot = 0
    
    #for each timestep
    for timestep in hist:
        if InitialTime <= timestep[0] <= FinalTime:
            tot += 1
            
            #for each complex size
            for complex_size in timestep[2]:
                
                #initilize this size + it's counting
                if complex_size not in plot_conv:
                    plot_conv.append(complex_size)
                    plot_count.append(timestep[1][timestep[2].index(complex_size)])
                
                #add size to the list
                else:
                    index = plot_conv.index(complex_size)
                    plot_count[index] += timestep[1][timestep[2].index(complex_size)]
        elif timestep[0] > FinalTime:
            break
    
    plot_count_mean = []
    
    for complex_count in plot_count:
        plot_count_mean.append(complex_count/tot)
    return plot_conv, plot_count_mean





import numpy as np
import matplotlib.pyplot as plt
import warnings
from .read_transition_matrix import read_transition_matrix
from ..file_managment.save_vars_to_file import save_vars_to_file

def associate_prob_symmetric(FileName: str, FileNum: int, InitialTime: float, FinalTime: float,
                             SpeciesName: str, DivideSize: int = 2, ShowFig: bool = True, SaveFig: bool = False, SaveVars: bool = False):
    """ Plots the probability of association between complexes of different sizes and other complexes of different sizes.

    Args:
        FileName (str): Path to the histogram data file (usually named as 'transition_matrix_time.dat') to be analyzed.
        FileNum (int): Number of the total input files. If multiple files are provided, their names should obey the naming rule listed below.
        InitialTime (float): The initial time that users desire to examine. The acceptable range should not be smaller than the starting time or exceed the ending time of simulation.
        FinalTime (float): The final time that users desire to examine. The acceptable range should not be smaller than the value of InitialTime or exceed the ending time of simulation.
        SpeciesName (str): The name of the species that users want to examine, which should also be identical to the name written in the input (.inp and .mol) files.
        DivideSize (int, optional): Value that distinguishes the size of the associate complex. Defaults to 2.
        ShowFig (bool, optional): Whether to show the plot. Defaults to True.
        SaveFig (bool, optional): Whether to save the plot as a '.png' file in the current directory. Defaults to False.
        SaveVars (bool, optional): If the variables are saved to a file. Defaults to false.

    Returns:
        A tuple of four lists containing: 
        - the sizes 
        - probabilities of association for complexes of size less than DivideSize
        - probabilities of association for complexes of size equal to DivideSize
        - probabilities of association for complexes of size greater than DivideSize
    
    Notes:
        If multiple input files are given, the output plot will be the average value of all files, and an error bar will also be included.
        Naming rule for input files: If single file is provided, the input file should be named as its original name (‘transition_matrix_time.dat’); if multiple files are provided, the name of input file should also include serial number as ‘transition_matrix_time_X.dat’ where X = 1,2,3,4,5…

    """
    
    warnings.filterwarnings('ignore')
    matrix_list = []
    
    #create name head/tail
    file_name_head = FileName.split('.')[0]
    file_name_tail = FileName.split('.')[1]
    
    #for each transition matrix file input
    for matrix_file_number in range(1, FileNum+1):
        
        #determining file name (if there are multiple or none)
        if FileNum == 1:
            temp_file_name = FileName
        else:
            temp_file_name = file_name_head + '_' + str(matrix_file_number) + '.' + file_name_tail
        
        #read transition matrix (t1_matrix = matrix of proteins at initial, tf...)
        ti_matrix, tf_matrix = read_transition_matrix(
            temp_file_name, SpeciesName, InitialTime, FinalTime)
        
        #get change in matrix
        matrix = tf_matrix - ti_matrix
        matrix_list.append(matrix)
    
    
    above = []
    equal = []
    below = []
    #for each file
    for matrix in matrix_list:
        
        #create lists of zeros, based on the number of columns
        above_temp = np.zeros(len(matrix_list[0][0]))
        equal_temp = np.zeros(len(matrix_list[0][0]))
        below_temp = np.zeros(len(matrix_list[0][0]))
        
        #goes through the matrix, and based on row/column of data it gets assigned to equal/above/below
        row = 0
        while row < len(matrix):
            column = 0
            while column < len(matrix[row]):
                if row > column:
                    if row - column == DivideSize:
                        equal_temp[column] += matrix[row][column]
                    elif row - column > DivideSize:
                        above_temp[column] += matrix[row][column]
                    else:
                        below_temp[column] += matrix[row][column]
                column += 1
            row += 1
        #add to main above/ect. vars
        above.append(above_temp)
        equal.append(equal_temp)
        below.append(below_temp)
    
    #for each column determine probability based on entire count in that column. 
    above_prob = []
    equal_prob = []
    below_prob = []
    for i in range(len(above)):
        above_prob_temp = []
        equal_prob_temp = []
        below_prob_temp =[]
        for j in range(len(above[0])):
            sum = above[i][j] + equal[i][j] + below[i][j]
            if sum != 0:
                above_prob_temp.append(above[i][j]/sum)
                equal_prob_temp.append(equal[i][j]/sum)
                below_prob_temp.append(below[i][j]/sum)
            else:
                above_prob_temp.append(np.nan)
                equal_prob_temp.append(np.nan)
                below_prob_temp.append(np.nan)
        above_prob.append(above_prob_temp)
        equal_prob.append(equal_prob_temp)
        below_prob.append(below_prob_temp)
    
    #TRANSPOSING TIME!!!!!!
    above_prob_rev = np.array(above_prob).transpose()
    equal_prob_rev = np.array(equal_prob).transpose()
    below_prob_rev = np.array(below_prob).transpose()
    
    #calculate means and stds devss
    mean_above = []
    mean_equal = []
    mean_below = []
    std_above = []
    std_equal = []
    std_below = []
   
   #for every column (across all files) determine mean / std
    if FileNum != 1: 
        for i in range(len(above_prob_rev)):
            if np.any(above_prob_rev[i] != above_prob_rev[i-1]) or i == 0:
                mean_above.append(np.nanmean(above_prob_rev[i]))
                std_above.append(np.nanstd(above_prob_rev[i]))
            else:
                mean_above.append(mean_above[i-1])
                std_above.append(std_above[i-1])
            
            if np.any(equal_prob_rev[i] != equal_prob_rev[i-1]) or i == 0:
                mean_equal.append(np.nanmean(equal_prob_rev[i]))
                std_equal.append(np.nanstd(equal_prob_rev[i]))
            else:
                mean_equal.append(mean_equal[i-1])
                std_equal.append(std_equal[i-1])

            if np.any(below_prob_rev[i] != below_prob_rev[i-1]) or i == 0:
                mean_below.append(np.nanmean(below_prob_rev[i]))
                std_below.append(np.nanstd(below_prob_rev[i]))
            else:
                mean_below.append(mean_below[i-1])        
                std_below.append(std_below[i-1])
    else:
        for i in range(len(above_prob_rev)):
            mean_above.append(above_prob_rev[i])
            mean_equal.append(equal_prob_rev[i])
            mean_below.append(below_prob_rev[i])

        
    n_list = list(range(1, 1 + len(matrix_list[0])))
        
    #output variables
    if SaveVars:
        save_vars_to_file({"cmplx_size":n_list,"mean_associate_probability":[mean_above, mean_equal, mean_below],"std":[std_above, std_equal, std_below]})

    #show figure!
    if ShowFig:
        errorbar_color_1 = '#c9e3f6'
        errorbar_color_2 = '#ffe7d2'
        errorbar_color_3 = '#d7f4d7'
        plt.plot(n_list, mean_above, 'C0')
        plt.plot(n_list, mean_equal, 'C1')
        plt.plot(n_list, mean_below, 'C2')
        if FileNum != 1:
            plt.errorbar(n_list, mean_above, yerr=std_above,
                         ecolor=errorbar_color_1, capsize=2)
            plt.errorbar(n_list, mean_equal, yerr=std_equal,
                         ecolor=errorbar_color_2, capsize=2)
            plt.errorbar(n_list, mean_below, yerr=std_below,
                         ecolor=errorbar_color_3, capsize=2)
        plt.legend(['Associate Size > ' + str(DivideSize), 'Associate Size = ' +
                    str(DivideSize), 'Associate Size < ' + str(DivideSize)])
        plt.xlabel('Number of ' + str(SpeciesName) + ' in Single Complex')
        plt.ylabel('Probability')
        plt.xticks(ticks=n_list)
        plt.title('Symmetric Association Probability')
        if SaveFig:
            plt.savefig('associate_probability_symmetric.png', dpi=500)
        plt.show()
    return n_list, [mean_above, mean_equal, mean_below], [std_above, std_equal, std_below]


